import gurobipy as gp
from gurobipy import GRB
import time
import numpy as np

def decode_binary_string(x, height, width):
    """
    Decode a binary string into a binary segmentation mask.

    Args:
        x (list): Binary string representing the segmentation.
        height (int): Height of the image.
        width (int): Width of the image.

    Returns:
        numpy.ndarray: Segmentation mask.
    """
    mask = np.zeros([height, width])
    for index, segment in enumerate(x):
        mask[index // width, index % width] = segment
    return mask

def QUBO_formulation(G, alpha=10, height=32):
    ''' 
    Function that provides the QUBO formulation for quantum and simulated annealing. It provides the linear and quadratic constraints.

    Args:
        G (networkx.Graph): Weighted graph encoding the input image
        alpha (float): hyperparameter controlling the importance of the smoothness constraint.

    Returns:
        linear (dict): linear constraints of QUBO model
        quadratic (dict): quadratic constraints of QUBO model
        problem_formulation_time (float): Problem formulation time
    '''
    # Problem formulation
    node_dict = {list(G.nodes())[i]:i for i in range(len(G.nodes()))}
    linear = {i:0 for i in range(len(node_dict))}
    quadratic1 = {(node_dict[i],node_dict[j]):0 for (i,j) in G.edges()}
    quadratic2 = {(i,i):0 for i in range(len(node_dict))}
    quadratic = {**quadratic1, **quadratic2}
    start_time = time.time()
    for edge in G.edges():
        node1, node2 = edge[0], edge[1]
        w = G.get_edge_data(node1, node2)['weight']
        node1_idx, node2_idx = node1[0]*height + node1[1], node2[0]*height + node2[1]
        # Min-cut constraints w_{ij}*x_i(1- x_j)
        linear[node1_idx] += w 
        linear[node2_idx] += w 
        quadratic[(node1_idx, node2_idx)] += -2*w
        # Smootheness constraints  S = alpha*w_{ij}(q- delta(xi, xj)) = alpha*w_{ij}(1-(xi + xj - 1)^2) = alpha*w_{ij}(2xi + 2xj -2xixj -xi^2 - xj^2)
        linear[node1_idx] += alpha*2*w 
        linear[node2_idx] += alpha*2*w 
        quadratic[(node1_idx, node2_idx)] += -alpha*2*w
        quadratic[(node1_idx, node1_idx)] += -alpha*w
        quadratic[(node2_idx, node2_idx)] += -alpha*w
    problem_formulation_time = time.time() - start_time
    return linear, quadratic, problem_formulation_time




def gurobi_qubo_solver(G, alpha=10, height=32, width=32):
    ''' 
    Gurobi solver.

    Args
        G (networkx.Graph): Weighted graph encoding the input image
        alpha (float): hyperparameter controlling the importance of the smoothness constraint.
        
    Returns:
        segmentation_mask (np.array): Segmentation mask
        gurobi_qubo_value (float): QUBO value of segmentation mask
    '''
    # Problem formulation
    model = gp.Model()
    n = len(G.nodes())
    x = model.addVars(n, vtype=GRB.BINARY)
    obj_expr = 0 
    model.update()
    for edge in G.edges():
        node1, node2 = edge[0], edge[1]
        w = G.get_edge_data(node1, node2)['weight']
        node1_idx, node2_idx = node1[0]*height + node1[1], node2[0]*height + node2[1]
        obj_expr += w*(x[node1_idx] + x[node2_idx] - 2*x[node1_idx]*x[node2_idx]) + alpha*w*(1-(x[node1_idx] + x[node2_idx] - 1)**2)
    model.setObjective(obj_expr)
    model.setParam('OutputFlag', 0)

    model.optimize()
    model.update()
    if model.status == GRB.OPTIMAL:
        solution = [int(x[i].X) for i in range(n)]
        binary_string = ''.join(str(bit) for bit in solution)
        gurobi_qubo_solution, gurobi_qubo_value = binary_string, model.objVal

        segmentation_mask = decode_binary_string(gurobi_qubo_solution, height, width)
        return segmentation_mask, gurobi_qubo_value
    else:
        print('No solution found')
        return None, None


import numpy as np
from qiskit import QuantumCircuit, QuantumRegister, ClassicalRegister
from qiskit.quantum_info import Statevector
from qiskit.converters import circuit_to_dag
from qiskit.visualization import circuit_drawer
from qiskit.circuit.tools.pi_check import pi_check
from qiskit.quantum_info import partial_trace, Operator, DensityMatrix, purity

from io import BytesIO
from PIL import Image

import matplotlib
import matplotlib.pyplot as plt
from matplotlib import gridspec
from matplotlib.patches import FancyArrowPatch, Circle, Ellipse
from mpl_toolkits.mplot3d import proj3d
from mpl_toolkits.mplot3d.art3d import Patch3D
from functools import reduce
from .custom import custom_style
import warnings


class Arrow3D(Patch3D, FancyArrowPatch):
    """
    A 3D arrow class for visualizing quantum states on a Bloch sphere or Q-sphere.

    This class extends `Patch3D` and `FancyArrowPatch` to create a 3D arrow
    that can be used to represent vectors in quantum state visualization.

    Attributes:
        xs (list): X-coordinates of the arrow.
        ys (list): Y-coordinates of the arrow.
        zs (list): Z-coordinates of the arrow.
        zdir (str): Direction of the arrow (default: "z").
    """
    __module__ = "mpl_toolkits.mplot3d.art3d"

    def __init__(self, xs, ys, zs, zdir="z", **kwargs):
        """
        Initializes the 3D arrow with the given coordinates and direction.

        Args:
            xs (list): X-coordinates of the arrow.
            ys (list): Y-coordinates of the arrow.
            zs (list): Z-coordinates of the arrow.
            zdir (str, optional): Direction of the arrow. Default is "z".
            **kwargs: Additional arguments for `FancyArrowPatch`.
        """
        FancyArrowPatch.__init__(self, (0, 0), (0, 0), **kwargs)
        self.set_3d_properties(tuple(zip(xs, ys)), zs, zdir)
        self._path2d = None

    def draw(self, renderer):
        """
        Renders the 3D arrow on the given Matplotlib renderer.

        Args:
            renderer (matplotlib.backend_bases.RendererBase): The renderer to draw the arrow.
        """
        xs3d, ys3d, zs3d = zip(*self._segment3d)
        x_s, y_s, _ = proj3d.proj_transform(xs3d, ys3d, zs3d, self.axes.M)
        self._path2d = matplotlib.path.Path(np.column_stack([x_s, y_s]))
        self.set_positions((x_s[0], y_s[0]), (x_s[1], y_s[1]))
        FancyArrowPatch.draw(self, renderer)

class QuantumVisualizer:
    def __init__(self, qc):
        """
        Initialize the StateSpaceVisualizer with a quantum circuit.

        Args:
            qc (QuantumCircuit): The input quantum circuit or statevector.
        """
        self.qc = qc
        self.display_math = False
        self.arrows = []  # Stores arrow connections
        if isinstance(qc, QuantumCircuit):
            self.num_qubits = qc.num_qubits
            self.vectors, self.gate_set = self.get_statevectors()

    @staticmethod
    def bit_string_index(s):
        """
        Return the index of a string of 0s and 1s.

        Args:
            s (string): bit string

        Returns:
            int: index of the string.
        """
        n = len(s)
        k = s.count("1")
        if s.count("0") != n - k:
            raise ValueError("s must be a string of 0 and 1")
        ones = [pos for pos, char in enumerate(s) if char == "1"]
        return QuantumVisualizer.lex_index(n, k, ones)

    @staticmethod
    def n_choose_k(n, k):
        """
        Return the number of combinations for n choose k.

        Args:
            n (int): the total number of options .
            k (int): The number of elements.

        Returns:
            int: binomial coefficient
        """
        if n == 0:
            return 0
        return reduce(lambda x, y: x * y[0] / y[1], zip(range(n - k + 1, n + 1), range(1, k + 1)), 1)
    
    @staticmethod
    def lex_index( n, k, lst):
        """
        Return  the lex index of a combination..

        Args:
            n (int): the total number of options .
            k (int): The number of elements.
            lst (list): list

        Returns:
            int: returns int index for lex order

        Raises:
            VisualizationError: if length of list is not equal to k
        """
        if len(lst) != k:
            raise VisualizationError("list should have length k")
        comb = list(map(lambda x: n - 1 - x, lst))
        dualm = sum(QuantumVisualizer.n_choose_k(comb[k - 1 - i], i + 1) for i in range(k))
        return int(dualm)
    
    @staticmethod
    def count_control_num(gate_name):
        """
        Count the number of consecutive 'c' characters in a gate name.

        Args:
            gate_name (str): The name of the gate.

        Returns:
            int: The number of consecutive 'c' characters.
        """
        return len(gate_name) - len(gate_name.lstrip('c'))


    def get_transformed_bitstrings(self, unitary, tol=1e-8):
        """
        Compute the transformation of all basis states under a unitary matrix and output the transformed bitstrings.

        Args:
            unitary (np.ndarray): The unitary matrix of size (2^n, 2^n).
            tol (float): Tolerance for considering a value as non-zero.

        Returns:
            dict: A dictionary mapping input bitstrings to their transformed bitstrings.
        """
        # Determine the number of qubits
        dim = unitary.shape[0]
        n = int(np.log2(dim))

        # Prepare a dictionary for the results
        transformations = {}
        for i in range(dim):
            # Input basis state as a bitstring
            input_bitstring = format(i, f"0{n}b")
            # Transformation of the basis state (column of the unitary matrix)
            transformed_state = unitary[:, i]
            # Identify the output bitstrings with significant amplitudes
            transformed_bitstrings = [
                format(idx, f"0{n}b") for idx, amp in enumerate(transformed_state)
                if np.abs(amp) > tol
            ]
            # Store the mapping
            transformations[input_bitstring] = transformed_bitstrings
        return transformations
    
    def get_statevectors(self):
        """
        Extract statevectors and gate information from the quantum circuit.

        Returns:
            tuple: (list of statevectors, gate set)
        """
        dag = circuit_to_dag(self.qc)
        state = Statevector.from_label('0' * self.qc.num_qubits)
        vectors = [np.array(state)]
        gate_set = [{}]

        # Create new registers with desired names
        new_qr = QuantumRegister(len(self.qc.qubits), 'q')  # Renamed quantum register
        new_cr = ClassicalRegister(len(self.qc.clbits), 'c') if self.qc.clbits else None  # Renamed classical register
        self.qc_barriers =  QuantumCircuit(new_qr, new_cr) if new_cr else QuantumCircuit(new_qr)
        
        for i, layer in enumerate(dag.layers()):
            # Create new registers with desired names
            layer_circuit = QuantumCircuit(new_qr, new_cr) if new_cr else QuantumCircuit(new_qr)
            
            nodes, qargs_list = [], []
            # Iterate over all operation nodes (gates) in the layer
            for node in layer['graph'].op_nodes():
                if node.name not in ['barrier', 'measure', 'reset', 'initialize']:
                    # Map the qubits and classical bits to the new registers
                    qargs = [new_qr[self.qc.qubits.index(q)] for q in node.qargs]
                    cargs = [new_cr[self.qc.clbits.index(c)] for c in node.cargs] if node.cargs else []
                    # Add the gate to the layer circuit and the full circuit with barriers
                    layer_circuit.append(node.op, qargs=qargs, cargs=cargs) 
                    self.qc_barriers.append(node.op, qargs=qargs, cargs=cargs)
                    nodes.append(node)
                    qargs_list.append(qargs)         
                elif node.name == 'initialize':
                    warnings.warn('Initialize operation not implemented for state space vizualization.')

            if len(nodes)>0:
                # Build circuit for gate
                flattened_qargs = [item for sublist in qargs_list for item in sublist]
                flattened_qargs = sorted(flattened_qargs, key=lambda qubit: qubit._index)
                qubits = [n._index for n in flattened_qargs]
                qc_u = QuantumCircuit(flattened_qargs)
                for j, node in enumerate(nodes):
                    qc_u.append(node.op, qargs=qargs_list[j])
                # Get unitary matrix
                U = Operator(qc_u).data
                # Get bitstrings transformations
                bitstring_transformations = self.get_transformed_bitstrings(U, tol=1e-8)
                gate_set.append({
                    'layer_id': i,
                    'num_qubits': len(qubits),
                    'name': [n.name for n in nodes],
                    'qubits': qubits,
                    'bitstring_transformations': bitstring_transformations
                })

                state = state.evolve(layer_circuit)
                vectors.append(np.array(state))
                self.qc_barriers.barrier() 

        return vectors, gate_set

    @staticmethod
    def hex_to_rgb(hex_str):
        """
        Convert a hexadecimal color string to an RGB tuple.

        Args:
            hex_str (str): The hexadecimal color string.

        Returns:
            np.ndarray: The RGB color as a numpy array.
        """
        return np.array([int(hex_str[i:i+2], 16) for i in range(1, 6, 2)]) / 255

    @staticmethod
    def color_range(c1, c2, n):
        """
        Generate a color gradient between two colors.

        Args:
            c1 (str): The starting color in hexadecimal.
            c2 (str): The ending color in hexadecimal.
            n (int): The number of gradient steps.

        Returns:
            list: A list of colors in hexadecimal format.
        """
        c1_rgb = QuantumVisualizer.hex_to_rgb(c1)
        c2_rgb = QuantumVisualizer.hex_to_rgb(c2)
        return ["#" + "".join(f"{int(round(val * 255)):02x}" for val in (c1_rgb * (1 - t) + c2_rgb * t)) for t in np.linspace(0, 1, n)]

    @staticmethod
    def get_colors(n):
        """
        Generate a full cyclic color map with gradients.

        Args:
            n (int): Total number of colors.

        Returns:
            list: A list of colors in hexadecimal format.
        """
        m = n // 4
        return (
             QuantumVisualizer.color_range("#55d397", "#bb6bd6", m) 
            + QuantumVisualizer.color_range("#bb6bd6", "#ffcc3e", m)
            + QuantumVisualizer.color_range("#ffcc3e", "#c1c5c9", m)
            + QuantumVisualizer.color_range("#c1c5c9", "#55d397", m)
        )

    @staticmethod
    def phase_to_rgb(complex_number, n=64, tol=1e-3, return_black = True):
        """
        Map a phase of a complex number to a color.

        Args:
            complex_number (complex): The input complex number.
            n (int): Number of colors in the gradient.

        Returns:
            str: RGB color in hexadecimal format.
        """
        
        if np.abs(complex_number)<tol and return_black:
            return 'black'
        colors = QuantumVisualizer.get_colors(n)
        offset_colors = colors[-n // 4 :] + colors[:-n // 4]
        complex_number = np.round(complex_number, 3)
        real_part = 0 if np.isclose(abs(complex_number.real), 0) else complex_number.real
        im_part = 0 if np.isclose(abs(complex_number.imag), 0) else complex_number.imag
        corrected_num = real_part + 1j*im_part
        angle = (np.angle(corrected_num) % (2 * np.pi)) / (2 * np.pi)
        return offset_colors[int(angle * n)]



    def draw_basis_states(self):
        """
        Draw basis states as text labels on the y-axis.
        """
        basis_states = [f"|{format(i, f'0{self.num_qubits}b')}⟩" for i in range(2 ** self.num_qubits)]
        font_size = max(8, 22 - 2 * self.num_qubits) 
        for i, state in enumerate(basis_states):
            plt.text(0.5 - 0.03*len(self.vectors), i / len(basis_states), state, ha='center', fontsize=font_size)

    def draw_probabilities(self, statevector, position, tol=1e-3):
        """
        Draw probability circles for a given statevector.

        Args:
            statevector (np.ndarray): The statevector.
            position (int): The x-axis position of the circles.
            ax ( matplotlib.axes): axes to draw the probabilities
            tol (float): Tolerance for non-zero probabilities.
        """
        y_coords = np.arange(len(statevector)) / len(statevector)
        for i, value in enumerate(statevector):
            radius = np.abs(value)
            color = self.phase_to_rgb(value, 64, tol) # Map the phase of the complex number to a color
            base_size = 2500 / self.num_qubits if self.num_qubits>1 else 1800
            size = radius**2 * base_size if radius >= tol else base_size/100
            plt.scatter(position + 1, y_coords[i], s=size, color=color, zorder=1) # Plot the statevector magnitude
            # If `display_math` is enabled, overlay a rectangle and text showing the value
            if self.display_math and radius>=tol:
                plt.scatter(position + 1, y_coords[i], s=base_size, color='#73BBEF',alpha=0.5, marker='s', zorder=2)
                # Format the value string based on the real and imaginary parts
                if np.abs(value.imag)<tol:
                    value_str = str(np.round(value.real,2))
                elif np.abs(value.real)<tol:
                    value_str = str(np.round(value.imag,2)) + 'i'
                else:
                    value_str = str(np.round(value.real,2)) + str(np.round(value.imag,2)) + 'i'
                font_size = max(8, int(18 - 1.5 * self.num_qubits)) 
                # Add the text in the center of the rectangle
                plt.text(position + 1, y_coords[i], value_str, color='black', ha='center', va='center', fontsize=font_size, zorder=3)


    @staticmethod
    def flip_bit(bitstring, i):
        """
        Flip the bit at position i in a binary string.

        Args:
            bitstring (str): The binary string.
            i (int): The position of the bit to flip (0-indexed).

        Returns:
            str: The binary string with the bit at position i flipped.
        """
        binary_list = list(bitstring)[::-1]
        binary_list[i] = '1' if binary_list[i] == '0' else '0'
        return ''.join(binary_list[::-1])

    def draw_arrow(self, vector, old_vector, position, tol=1e-3):
        """
        Draw arrows connecting states between layers.

        Args:
            vector (np.ndarray): Current statevector (after the gate is applied).
            old_vector (np.ndarray): Previous statevector (before the gate is applied).
            position (int): The layer index (x-axis position in the visualization).
            tol (float): Tolerance for non-zero probabilities.
        """       
        # Generate normalized y-coordinates 
        y_coords = np.arange(len(vector)) / len(vector)
        
        # Generate binary labels for all basis states
        basis_states = [f"{format(i, f'0{self.num_qubits}b')}" for i in range(2 ** self.num_qubits)]
        idxs = np.arange(len(vector))  # Indexes of all basis states

        # Identify basis states with non-zero probabilities in the old and new vectors
        old_nonzero_states = np.array(basis_states)[idxs[np.abs(old_vector) > tol]].tolist()
        new_nonzero_states = np.array(basis_states)[idxs[np.abs(vector) > tol]].tolist()

        gate_info =  self.gate_set[position]
        for bitstr in old_nonzero_states:
            # Find the mapping frm the old vector basis state to the new one
            reduced_bitstring = "".join([bitstr[::-1][q] for q in gate_info['qubits'][::-1]]) # Reversing order to keep convention of basis states 
            reduced_bitstring_complementary = "".join([bitstr[::-1][q] for q in list(range(self.qc.num_qubits))[::-1] if q not in gate_info['qubits']])
            child_bitstr = []
            for new_bitstr_short in gate_info['bitstring_transformations'][reduced_bitstring]:
                new_bitstr = ""
                i_r = 0
                i_o = 0
                for n in list(range(self.qc.num_qubits))[::-1]:
                    if n in gate_info['qubits']:
                        new_bitstr += new_bitstr_short[i_r]
                        i_r+=1
                    else:
                        new_bitstr += reduced_bitstring_complementary[i_o] 
                        i_o+=1
                new_bitstr = new_bitstr  
                child_bitstr.append(new_bitstr)    
            child_bitstr = [b for b in child_bitstr if b in new_nonzero_states]

            # Draw arrows from parent states to the current state
            for child in child_bitstr:                
                bitstr_idx = basis_states.index(bitstr)  # Index of the current state
                child_bitstr_idx = basis_states.index(child)  # Index of the child state
                x, y = position, y_coords[bitstr_idx]  # Start point of the arrow
                dx, dy = 1, y_coords[child_bitstr_idx] - y  
                color = self.phase_to_rgb(vector[child_bitstr_idx])  # Determine the arrow color based on the state's phase
                plt.arrow(
                    x, y, dx, dy, color=color, linewidth=max(3, 3 + (7 - 3) * (5 - self.num_qubits) / 4), head_length=0, alpha=0.5  
                )

    def draw_state_space(self,  plot_circuit=False, plot_phase_disks = False, display_math = False, filename = None):
        """
        Visualize the quantum circuit's evolution. Optionally include the circuit diagram.

        Args:
            plot_circuit (bool): If True, include the quantum circuit diagram in the figure.
            plt_phase_disks (bool): If True, include the diagram of phase disks in the figure.
            display_math (bool): If True, include the value of the statevectors in the figure.
            filename (str): File name to save the figure.

        Returns:
            matplotlib.pyplot.figure: figure of the state space        
        """
        self.display_math = display_math
        base_size = 1 / (4 * self.num_qubits)
        figure_size = (
            max(8, len(self.vectors) + 2),
            max(4, (1.7) ** self.num_qubits + 1.7*self.num_qubits if plot_phase_disks else max(4, (1.7) ** self.num_qubits)),
        )  # Adjust figure size dynamically

        if plot_circuit:
            # Save the Qiskit circuit with barriers as an image
            dpi = 100
            circuit_image = BytesIO()
            circuit_drawer(
                self.qc_barriers, output="mpl", style=custom_style
            ).savefig(circuit_image, format="png", bbox_inches="tight", dpi=dpi, transparent = True)
            circuit_image.seek(0)
            circuit_img = Image.open(circuit_image)

            # Get the size of the circuit image in pixels
            img_width, img_height = circuit_img.size

            # Calculate the height of the image in inches
            image_fig_height = img_height / dpi

            # Create the figure with two subplots
            fig, (ax_top, ax_bottom) = plt.subplots(
                2, 1,
                figsize=(figure_size[0], figure_size[1] + image_fig_height * 1.6 ),
                gridspec_kw={"height_ratios": [image_fig_height * 1.6, figure_size[1] ]},
            )
            bottom_ratio = figure_size[1] / (figure_size[1] + image_fig_height * 1.6)

            # Aspect ratio and extent for the top image
            aspect_ratio = img_width / img_height
            ax_top.imshow(circuit_img, extent=[0, aspect_ratio, 0, 1])
            ax_top.axis("off")
            ax_top.set_xlim(
                -1.1 / len(self.vectors), aspect_ratio + 1.1 / len(self.vectors)
            )
            ax_top.set_ylim(0, 1)

            # Set the bottom axis for state space visualization
            ax = ax_bottom
        else:
            bottom_ratio = 1
            # Create a single-plot figure if no circuit diagram is included
            fig, ax = plt.subplots(figsize=figure_size)


        # Draw the state space visualization
        self.draw_basis_states()
        for i, vector in enumerate(self.vectors):
            if i > 0:
                self.draw_arrow(vector, self.vectors[i - 1], i)
                self.draw_probabilities(self.vectors[i - 1], i - 1)
        self.draw_probabilities(self.vectors[-1], len(self.vectors) - 1)

        # Plot the phase disks
        if plot_phase_disks:
            # Set up the state space axis
            ax.set_xlim(-0.2- 0.03*len(self.vectors), len(self.vectors) + 1)#(-0.2, len(self.vectors) + 1)
            ax.set_ylim(-base_size -(self.num_qubits + 2) / (2**self.num_qubits), 1)

            ax.axhline(y = -(1) / (2**self.num_qubits), color='black', linestyle='--')
            for i, vector in enumerate(self.vectors):
                for qubit in range(self.num_qubits):
                    pos_x = i + 1
                    pos_y = -(qubit + 2) / (2**self.num_qubits)
                    # Get partial trace
                    v = Statevector(vector)
                    partial_state = partial_trace(v, [i for i in range(self.num_qubits) if i!=qubit]).data
                    self._draw_disk(partial_state, qubit, radius = 0.45/np.sqrt(self.num_qubits), ax = ax, 
                                    pos_x=pos_x, pos_y=pos_y, plot_text = True if i == 0 else False, pos_label=0.5 - 0.05*len(self.vectors),
                                    bottom_ratio=bottom_ratio)         
        else:
            ax.set_xlim(-0.2 - 0.03*len(self.vectors), len(self.vectors) + 1)
            ax.set_ylim(-base_size, 1)
        ax.axis("off")

        if plot_circuit:
            plt.subplots_adjust(hspace=0.0)
        if filename:
            fig.savefig(filename, transparent = True)
        plt.close() 
        return fig
    
    @staticmethod
    def statevector_to_angles(phi):
        """
        Convert a quantum statevector into spherical coordinates (theta, phi) for Bloch sphere representation.

        Args:
            phi (ndarray): A 2-element array-like object representing a single-qubit statevector.

        Returns:
            theta (float): Polar angle (0 to pi).
            phi (float): Azimuthal angle (0 to 2*pi).
        """
        alpha, beta  = phi[0], phi[1]
        # Calculate theta, phi
        theta = 2 * np.arccos(np.abs(alpha))
        phi = np.angle(beta) - np.angle(alpha)
        return theta, phi

    def draw_bloch_sphere(self, theta=None, phi=None, ax = None, filename = None):
        """
        Visualize a single-qubit quantum state on the Bloch sphere.

        Args:
            theta (float): Polar angle for the Bloch vector (default 0).
            phi (float): Azimuthal angle for the Bloch vector (default 0).
            ax (matplotlib.axes): A Matplotlib axes to draw on (optional).
            filename (str): Path to save the plot (optional).

        Returns:
            matplotlib.figure.Figure: The Matplotlib figure object (if `fig` is not provided).
        """
        vector = Statevector(self.qc).data
        num_qubits = int(np.log2(len(vector)))
        if num_qubits!=1:
            raise ValueError("Input is not a single-qubit quantum state.")

        # Create Figure
        if ax is None:
            return_fig = True
            fig = plt.figure(figsize=(5,5))
            ax = fig.add_subplot(111, projection='3d')
        else:
            fig = ax.get_figure()
            return_fig = False
            
        # Draw a sphere
        ax.axes.set_xlim3d(-1.0, 1.0)
        ax.axes.set_ylim3d(-1.0, 1.0)
        ax.axes.set_zlim3d(-1.0, 1.0)
        ax.axes.grid(False)
        ax.view_init(elev=5, azim=275)

        # Force aspect ratio
        if hasattr(ax.axes, "set_box_aspect"):
            ax.axes.set_box_aspect((1, 1, 1))

        # Plot semi-transparent sphere
        u = np.linspace(0, 2 * np.pi, 25)
        v = np.linspace(0, np.pi, 25)
        x = np.outer(np.cos(u), np.sin(v))
        y = np.outer(np.sin(u), np.sin(v))
        z = np.outer(np.ones(np.size(u)), np.cos(v))
        ls = matplotlib.colors.LightSource(azdeg=0, altdeg=65)
        rgb = ls.shade(z, plt.cm.Blues)
        ax.plot_surface(
            x, y, z, rstride=3, cstride=3, color="#73BBEE", facecolors = rgb,  alpha=0.1, linewidth=0.3,
        )
        # Get rid of the panes
        ax.xaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
        ax.yaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
        ax.zaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))

        # Get rid of the spines
        ax.xaxis.line.set_color((1.0, 1.0, 1.0, 0.0))
        ax.yaxis.line.set_color((1.0, 1.0, 1.0, 0.0))
        ax.zaxis.line.set_color((1.0, 1.0, 1.0, 0.0))

        # Get rid of the ticks
        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_zticks([])
        if theta is None or phi is None:
            theta, phi = QuantumVisualizer.statevector_to_angles(vector)
        # Draw the state vector
        x_arrow = np.sin(theta) * np.cos(phi)
        y_arrow = np.sin(theta) * np.sin(phi)
        z_arrow = np.cos(theta)

        ax.plot([0, x_arrow], [0, y_arrow], [0, z_arrow], color='r', linewidth=2)
        ax.scatter([x_arrow], [y_arrow], [z_arrow], color='r', s=100)
        
        # Set the axes properties
        ax.set_xlim([-0.73, 0.73])
        ax.set_ylim([-0.73, 0.73])
        ax.set_zlim([-0.73, 0.73])

        # Draw the axes
        ax.quiver(0, 0, 0, 1, 0, 0, color='k', linewidth=1, arrow_length_ratio=0.1)
        ax.quiver(0, 0, 0, 0, 1, 0, color='k', linewidth=1, arrow_length_ratio=0.1)
        ax.quiver(0, 0, 0, 0, 0, 1, color='k', linewidth=1, arrow_length_ratio=0.1)
        
        # Label the axes
        ax.text(1.05, 0, 0, 'X', color='k', fontsize=14)
        ax.text(0, 1.05, 0, 'Y', color='k', fontsize=14)
        ax.text(-0.03, 0, 1.08, 'Z', color='k', fontsize=14)

        ax.set_title(f'Bloch Sphere (θ={theta:.2f}, φ={phi:.2f})', pad = -60)
        
        if filename:
            fig.savefig(filename, transparent = True)
        if return_fig:
            plt.close() 
            return fig

    
    def draw_qsphere(self, figsize=None, ax=None,
                    show_state_labels=True, show_state_phases=False,
                    use_degrees=False, tol = 1e-3,filename=None):
        
        """
        Plot the qsphere representation of a quantum state.
        Here, the size of the points is proportional to the probability
        of the corresponding term in the state and the color represents
        the phase.

        Args:
            figsize (tuple): Figure size in inches.
            ax (matplotlib.axes.Axes): An optional Axes object to be used for
                the visualization output. If none is specified a new matplotlib
                Figure will be created and used. Additionally, if specified there
                will be no returned Figure since it is redundant.
            show_state_labels (bool): An optional boolean indicating whether to
                show labels for each basis state.
            show_state_phases (bool): An optional boolean indicating whether to
                show the phase for each basis state.
            use_degrees (bool): An optional boolean indicating whether to use
                radians or degrees for the phase values in the plot.
            tol (float): tolerance to show basis state on the sphere
            filename (str): File to save the plot if desired

        Returns:
            Figure: A matplotlib figure instance if the ``ax`` kwarg is not set

        """
        # Get quantum statevector
        state = Statevector(self.qc).data
        num_qubits = int(np.log2(len(state)))
        if num_qubits<2:
            raise ValueError("Input is not a multi-qubit quantum state.")

        # Create Figure
        if figsize is None:
            figsize = (5, 5)

        if ax is None:
            return_fig = True
            fig = plt.figure(figsize=figsize)
            ax = fig.add_subplot(111, projection="3d")
        else:
            return_fig = False
            fig = ax.get_figure()
        
        ax.axes.set_xlim3d(-1.0, 1.0)
        ax.axes.set_ylim3d(-1.0, 1.0)
        ax.axes.set_zlim3d(-1.0, 1.0)
        ax.axes.grid(False)
        ax.view_init(elev=5, azim=275)

        # Force aspect ratio
        if hasattr(ax.axes, "set_box_aspect"):
            ax.axes.set_box_aspect((1, 1, 1))

        # Plot semi-transparent sphere
        u = np.linspace(0, 2 * np.pi, 25)
        v = np.linspace(0, np.pi, 25)
        x = np.outer(np.cos(u), np.sin(v))
        y = np.outer(np.sin(u), np.sin(v))
        z = np.outer(np.ones(np.size(u)), np.cos(v))
        ls = matplotlib.colors.LightSource(azdeg=0, altdeg=65)
        rgb = ls.shade(z, plt.cm.Blues)
        ax.plot_surface(
            x, y, z, rstride=3, cstride=3, color="#73BBEE", facecolors = rgb,  alpha=0.1, linewidth=0.05,
        )
        # Get rid of the panes
        ax.xaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
        ax.yaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))
        ax.zaxis.set_pane_color((1.0, 1.0, 1.0, 0.0))

        # Get rid of the spines
        ax.xaxis.line.set_color((1.0, 1.0, 1.0, 0.0))
        ax.yaxis.line.set_color((1.0, 1.0, 1.0, 0.0))
        ax.zaxis.line.set_color((1.0, 1.0, 1.0, 0.0))

        # Get rid of the ticks
        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_zticks([])

        # Set the axes properties
        ax.set_xlim([-0.93, 0.93])
        ax.set_ylim([-0.93, 0.93])
        ax.set_zlim([-0.93, 0.93])
        ax.set_title(f'Q-sphere', pad = -60)

        # Add center point
        ax.plot([0], [0], [0],
                markerfacecolor="#73BBEE", markeredgecolor="#73BBEE",
                marker="o",markersize=3, alpha=1)
        
        # Add weight lines
        for weight in range(num_qubits + 1):
            theta = np.linspace(-2 * np.pi, 2 * np.pi, 100)
            z = -2 * weight / num_qubits + 1
            r = np.sqrt(1 - z**2)
            x = r * np.cos(theta)
            y = r * np.sin(theta)
            ax.plot(x, y, z, color="#73BBEE", lw=1.5, ls=":", alpha=1)


        idxs = np.arange(len(state)) # indices quantum state in decimal form
        non_zero_probs = idxs[np.abs(state) > tol]
        for id in non_zero_probs:
            # Get basis state
            binstr = bin(id)[2:].zfill(num_qubits)
            weight = binstr.count("1")
            
            # Get position of basis state in the sphere
            number_of_divisions = QuantumVisualizer.n_choose_k(num_qubits, weight)
            weight_order = QuantumVisualizer.bit_string_index(binstr)
            angle = (float(weight) / num_qubits) * (np.pi * 2) + (
                weight_order * 2 * (np.pi / number_of_divisions))
            
            # Remove global phase
            if (weight > num_qubits / 2) or ( (weight == num_qubits / 2) and (weight_order >= number_of_divisions / 2)):
                angle = np.pi - angle - (2 * np.pi / number_of_divisions)

            # Get cartesian coordinates
            zvalue = -2 * weight / num_qubits + 1
            xvalue = np.sqrt(1 - zvalue**2) * np.cos(angle)
            yvalue = np.sqrt(1 - zvalue**2) * np.sin(angle)

            # Get size of ball (probability)
            prob = np.abs(state[id])**2
            # Get color of the ball (complex angle)
            color = QuantumVisualizer.phase_to_rgb(state[id])

            # Plot state labels and angles
            if show_state_labels:
                # Transform to spherical coordinates for text
                angle_theta, rprime = np.arctan2(np.sqrt(1 - zvalue**2), zvalue), 1.35
                xvalue_text = rprime * np.sin(angle_theta) * np.cos(angle)
                yvalue_text = rprime * np.sin(angle_theta) * np.sin(angle)
                zvalue_text = rprime * np.cos(angle_theta)
                # State label
                element_text = "$\\vert" + binstr + "\\rangle$"
                # Add phases to label in degrees or radians
                if show_state_phases:
                    element_angle = np.angle(state[id])
                    if use_degrees:
                        element_text += "\n$%.1f^\\circ$" % (element_angle * 180 / np.pi)
                    else:
                        element_angle = pi_check(element_angle, ndigits=3).replace("pi", "\\pi")
                        element_text += "\n$%s$" % (element_angle)
                # Add label to figure
                ax.text( xvalue_text, yvalue_text, zvalue_text, element_text,
                        ha="center", va="center", size=12 )
            
            # Add ball to figure
            ax.plot([xvalue],[yvalue],[zvalue],
                    markerfacecolor=color, markeredgecolor=color,
                    marker="o", markersize=np.sqrt(prob) * 50, )
            
            # Add arrow to figure
            a = Arrow3D([0, xvalue], [0, yvalue], [0, zvalue],
                        mutation_scale=20,alpha=1, arrowstyle="-",
                        color=color, lw=2)
            ax.add_artist(a)

        
        if filename:
            fig.savefig(filename, transparent = True)
        if return_fig:
            plt.close() 
            return fig
        
    @staticmethod
    def draw_color_legend(figsize=None, ax=None, n = 64, use_degrees=False, filename=None):
        """
        Plot the color legend for the qsphere and state space.

        Args:
            figsize (tuple): Figure size in inches.
            ax (matplotlib.axes.Axes): An optional Axes object to be used for
                the visualization output. If none is specified a new matplotlib
                Figure will be created and used. Additionally, if specified there
                will be no returned Figure since it is redundant.
            n (int): Number of divisions of the color legend.
            use_degrees (bool): An optional boolean indicating whether to use
            degrees or complex numbers.
            filename (str): File to save the plot if desired

        Returns:
            Figure: A matplotlib figure instance if the ``ax`` kwarg is not set

        """
        if figsize is None:
            figsize = (5, 5)

        if ax is None:
            return_fig = True
            fig = plt.figure(figsize=figsize)
            ax = fig.add_subplot()
        else:
            return_fig = False
            fig = ax.get_figure()

        # Get rid of the ticks
        ax.set_xticks([])
        ax.set_yticks([])

        theta = np.ones(n)
        colors = QuantumVisualizer.get_colors(n)

        # Draw pie
        ax.pie(theta, colors=colors[47:] + colors[: 47], radius=0.75)
        ax.add_artist(Circle((0, 0), 0.5, color="white", zorder=1))
        ax.annotate("", xy=(0, 0), xytext=(0.9, 0), arrowprops=dict(arrowstyle="<-", lw=2.5))
        ax.annotate("", xy=(0, 0), xytext=(0, 0.9), arrowprops=dict(arrowstyle="<-", lw=2.5))
        ax.annotate("", xy=(0, 0), xytext=(-0.9, 0), arrowprops=dict(arrowstyle="<-", lw=2.5))
        ax.annotate("", xy=(0, 0), xytext=(0, -0.9), arrowprops=dict(arrowstyle="<-", lw=2.5))
        offset = 0.95  # since radius of sphere is one.

        if use_degrees:
            labels = [ "0º", "90º", "180º", "270º"]
        else:
            labels = ["$1$", "$i$", "$-1$", "$-i$"]

        # Write labels
        ax.text( offset, 0, labels[0], horizontalalignment="center", verticalalignment="center", fontsize=18)
        ax.text( 0, offset, labels[1], horizontalalignment="center", verticalalignment="center", fontsize=18)
        ax.text(-offset*1.05, 0, labels[2], horizontalalignment="center", verticalalignment="center", fontsize=18)
        ax.text( 0, -offset, labels[3], horizontalalignment="center", verticalalignment="center", fontsize=18)

        ax.set_xlim(-offset*1.3, offset*1.3)
        ax.set_ylim(-offset*1.3, offset*1.3)
        ax.set_title(f'Color legend', pad = -60)

        if filename:
            fig.savefig(filename, transparent = True)
        if return_fig:
            plt.close() 
            return fig

    @staticmethod
    def height_for_probability(radius, prob, total_area):
        """
        Calculate the height (h) of a circular segment corresponding to a given probability.

        Args:
            radius (float): Radius of the circle.
            prob (float): Desired probability (fraction of the circle's area to fill).
            total_area (float): Total area of the circle.

        Returns:
            float: Height (h) of the segment.
        """
        target_area = prob * total_area  # Target area to fill
        h_low, h_high = 0, 2 * radius   # Binary search bounds

        while h_high - h_low > 1e-6:  # Precision tolerance
            h_mid = (h_low + h_high) / 2
            area = QuantumVisualizer.area_segment(radius, h_mid)
            if area < target_area:
                h_low = h_mid
            else:
                h_high = h_mid
        return h_low
    
    @staticmethod
    def area_segment(r, h):
        """
        Calculate the area of a circular segment.

        Args:
            r (float): The radius of the circle.
            h (float): The height of the segment (distance from the chord to the arc).

        Returns:
            float: The area of the circular segment.
        """
        return r*r*np.arccos((r-h)/r) - (r-h)*np.sqrt(2*r*h - h*h)


    def _draw_disk(self, partial_state, qubit, ax, radius = 1, pos_x=0, pos_y = None,
                    plot_text = False, pos_label = 0, equal_aspect_ratio = False, bottom_ratio=1,
                    ):
        """ 
        Draw phase disks ar a certain position.

        Args:
            partial_state (np.arrray): Density matrix of reduced state
            qubit (int): Qubit which corresponds to this phase disk
            ax (matplotlib.axes): Axis to draw phase disks
            radius (float): Radius of the disk
            pos_x (float): position on the x axis to plot the phase disks
            pos_y (float): position on the y axis to plot the phase disks
            plot_text (bool): If True, plot the qubit label
            equal_aspect_ratio (bool): If True, ensure equal aspect ratio
            bottom_ratio (float): Used to calculate the aspect ratio
        """

        if equal_aspect_ratio:
            x_aspect = 1
            y_aspect = 1
            ax.set_aspect('equal')
        else:
            # Adjust to axis dimensions
            # Get the axis limits
            fig = ax.get_figure()
            xlim = ax.get_xlim()
            ylim = ax.get_ylim()

            # Calculate the aspect ratio of the axis scaling
            x_aspect = (xlim[1] - xlim[0]) / fig.get_size_inches()[0]  # Data-to-figure ratio (x)
            y_aspect = (ylim[1] - ylim[0]) / (fig.get_size_inches()[1]*bottom_ratio)  # Data-to-figure ratio (y)
           
        # Get radius and color for phase disk
        prob = partial_state[0,0].real
        color = QuantumVisualizer.phase_to_rgb(partial_state[1,0], return_black = False)

        total_area = np.pi*radius**2
        # Calculate the height of the segment to fill
        h_cutoff = QuantumVisualizer.height_for_probability(radius, prob, total_area)

        # Create grid for circle
        circle_res = 300
        y = np.linspace(-radius, radius, circle_res)
        x = np.sqrt(radius**2 - y**2)

        # Find the points up to the cutoff height
        cutoff_idx = np.where(y + radius <= h_cutoff)[0]
        y_cutoff = y[cutoff_idx]
        x_cutoff = x[cutoff_idx]
       
        ellipse = Ellipse((pos_x, pos_y), width=2 * radius * x_aspect, height=2 * radius * y_aspect, edgecolor=color, fill=False, linewidth=7)
        ax.add_patch(ellipse)

        # Create the polygon for the filled segment
        filled_y = np.concatenate([y_cutoff*y_aspect + pos_y, y_cutoff[::-1]*y_aspect + pos_y])
        filled_x = np.concatenate([-x_cutoff*x_aspect + pos_x, x_cutoff[::-1]*x_aspect + pos_x])

        # Plot the filled circle segment
        ax.fill(filled_x, filled_y, color=color, alpha=1)

        
        # Draw purity circle
        pur = purity(partial_state).real
        ellipse = Ellipse((pos_x, pos_y), width=2 * radius * pur * x_aspect, height=2 * radius* pur * y_aspect, edgecolor='black', fill=False, linewidth=2)
        ax.add_patch(ellipse)
        
        # Draw the circle outline   
        if plot_text: 
            ax.text(pos_label, pos_y, f'$q_{{{qubit}}}$', fontsize=(20 - self.num_qubits))

        if self.display_math:
            # Format the value string
            value_str = str(int(prob*100)) + '%'
            font_size = max(8, int(15 - 1.5 * self.num_qubits)) 
            # Add the text in the center of the rectangle
            plt.text(pos_x, pos_y, value_str, color='black', ha='center', va='center', fontsize=font_size, zorder=3)
        

    
    def draw_phase_disks(self, ax=None, filename=None, display_math = False):
        """
        Plot the color legend for the qsphere and state space.

        Args:
            ax (matplotlib.axes.Axes): An optional Axes object to be used for
                the visualization output. If none is specified a new matplotlib
                Figure will be created and used. Additionally, if specified there
                will be no returned Figure since it is redundant.
            filename (str): File to save the plot if desired
            display_math (bool): If True, show the fullness of the disk

        Returns:
            Figure: A matplotlib figure instance if the ``ax`` kwarg is not set
        """
        self.display_math = display_math
        if ax is None:
            return_fig = True
            figsize = (3,max(3, self.num_qubits/2))
            fig, ax = plt.subplots(1,1,figsize=figsize)
        else:
            return_fig = False
            fig = ax.get_figure()
        
        for qubit in range(self.num_qubits):
            # Get partial trace
            v = Statevector(self.qc)
            if self.num_qubits==1:
                partial_state = DensityMatrix(v).data
            else:
                partial_state = partial_trace(v, [i for i in range(self.num_qubits) if i!=qubit]).data
            self._draw_disk(partial_state, qubit, ax = ax, radius = 1, pos_x=0, pos_y = 2.5*self.num_qubits - 2.5*qubit,
                            plot_text = True, pos_label = - 2.2, equal_aspect_ratio= True)
                   
        ax.set_title(f'Phase disks', pad = -60)

        # Set limits and aspect ratio
        ax.set_xlim(-3, 2)
        ax.set_ylim(1, 2.5*(self.qc.num_qubits) + 2 )
        ax.axis('off')

        if filename:
            fig.savefig(filename, transparent = True)
        if return_fig:
            plt.close() 
            return fig
        
    def draw_geometric_plots(self, filename = None, display_math = False):
        """
        Create and visualize geometric representations of a quantum state, 
        including phase disks, the Q-sphere or Bloch sphere, and a color legend.

        Args:
            filename (str, optional): If provided, saves the generated plot 
                to the specified file. Defaults to None.
            display_math (bool): If True, show the fullness of the disks

        Returns:
            matplotlib.figure.Figure: The created figure containing the plots.
        """
        
        # Define figure size dynamically based on the number of qubits
        figsize = (5 + 5 + 5, max(5, self.num_qubits/2))
        fig = plt.figure(figsize=figsize)

        # Create a GridSpec layout with 3 columns
        spec = gridspec.GridSpec(nrows=1, ncols=3, figure=fig, width_ratios=[max(3, self.num_qubits/2), 5, 5])
        spec.update(wspace=-0.3)
        
        # Phase disk
        phase_disk_ax = fig.add_subplot(spec[0])
        self.draw_phase_disks(ax = phase_disk_ax, display_math = display_math)

        if self.num_qubits>1:
            # Q-sphere
            q_sphere_ax = fig.add_subplot(spec[1], projection='3d')
            self.draw_qsphere(ax = q_sphere_ax, show_state_labels=True)
        else:
            # Bloch-sphere
            bloch_sphere_ax = fig.add_subplot(spec[1], projection='3d')
            self.draw_bloch_sphere(ax = bloch_sphere_ax)
            
        # Color legend
        color_legend_ax = fig.add_subplot(spec[2])
        self.draw_color_legend(ax = color_legend_ax)


        # Save or return
        if filename:
            fig.savefig(filename, transparent = True)
        plt.close() 
        return fig
            


import numpy as np
from PIL import Image
from numpy.lib.stride_tricks import sliding_window_view
from scipy.signal import find_peaks


class QuantumInspiredImageProcessor:
    """
    Applies a quantum-inspired image filter.

    This transformation weighs each pixel's original intensity by a relative measure 
    of pairwise intensity difference versus total neighborhood contribution.

    **Class Parameters**:
        - **mu (float):** Steepness factor. Higher values lead to sharper curves, 
          whereas lower values provide smoother transitions.
        - **neighbor_size (int):** Neighborhood size for intensity computations.
        - **percentile (int):** Percentile to determine cluster values w.
        - **max_iter (int):** Maximum iterations for transformation.
        - **threshold (float):** Convergence threshold based on Mean Absolute Difference.
        - **L (int):** Number of cluster levels.

    **Example Parameters**:
        - mu = 0.4
        - neighbor_size = 3
        - percentile = 50
        - max_iter = 10
        - threshold = 1e-5
        - L = 12
    """

    def __init__(self, mu=0.40, neighbor_size = 3, percentile=95, max_iter=10, threshold=0.0001, L=8):
        
        self.mu = mu
        self.neighbor_size = neighbor_size
        self.percentile = percentile
        self.max_iter = max_iter
        self.threshold = threshold
        self.L = L
        self.info = {}

    def check_convergence(self, w1, w2):
        '''
        Check convergence of the quantum-inspired model.

        If the Mean Absolute Difference (MAD) of two consecutive matrices 
        is smaller than the threshold, the method has converged.

        Args:
            w1 (np.array): Previous image matrix.
            w2 (np.array): Current image matrix.

        Returns:
            bool: True if the model has converged, False otherwise.
            float: Mean Absolute Difference.
        '''
        mae = np.mean(np.abs(w1- w2))
        if mae < self.threshold:
            print('Final Mean Absolute Difference: {:0.4f}'.format(mae))
            return True, mae
        return False, mae

    def sum_neighbours(self, image):
        '''
        Computes the sum of the neighborhood of each pixel.

        Args:
            image (np.array): Input image.

        Returns:
            np.array: Sum of neighboring pixels for each pixel.
        '''
        # 1. Calculate image shape
        ht,_ = image.shape
        # 2. Calculate necessary padding
        pad = int(np.ceil((np.ceil(ht/self.neighbor_size)*self.neighbor_size - ht)/2))
        if pad>0: # Pad image with 0s if the size of the image is not a multiple of the neighbouring size
            image = np.pad(image, (pad,pad), mode='constant', constant_values=(0,0))
        # 3. Create a window view of the image
        window_view = sliding_window_view(image, (self.neighbor_size, self.neighbor_size))
        # 4. Sum the elements of the view
        sum_neighbourhood = np.sum(window_view, axis=(2,3))
        return sum_neighbourhood

    def transformation(self, I, cluster):
        """
        Applies a quantum-inspired transformation.

        Args:
            I (np.array): Input image.
            cluster (np.array): Array containing the cluster values [w0, w1, ..., wn].

        Returns:
            np.array: Quantum-inspired image transformation (one iteration).

            np.array: Alpha values calculated as `1 - (I_{i+p,j+q} - I_{ij})`.
        """
        I[I<0] = 0
        I[I>1] = 1
        ht, wt = I.shape[1], I.shape[0] 
        # 1. Calculate the sum of intensities of the 3x3 window for each pixel
        S = self.sum_neighbours(I)

        # Select cluster
        cluster = np.array(cluster)
        d = cluster.shape[0]
        # f is the quantum-inspired activation function f = sum( 1/(lamb + e^-mu(x-S)) ) 
        f = np.zeros((ht, wt))
        Alpha = np.zeros((ht, wt))
        idx = np.arange(d)
        for i in range(ht):
            for j in range(wt):
                p_vals = np.clip(np.arange(i - 1, i + 2), 0, ht - 1) # Indices [i-1, i, i+1] padding with 0s at the edges
                q_vals = np.clip(np.arange(j - 1, j + 2), 0, wt - 1) # Indices [j-1, j, j+1] padding with 0s at the edges
                # Find the index where I[i,j] >= cluster[k] and I[i,j] <= cluster[k + 1]
                if len(idx[I[i,j]>=cluster])==0:
                    print(I[i,j], cluster)
                k = min(idx[I[i,j]>=cluster][-1], d-2)
                if k+1 >=len(cluster):
                    lamb = S[i, j] / (1. - cluster[k])
                else:
                    lamb = S[i, j] / (cluster[k+1] - cluster[k])            
                sum_ = 0.
                alpha_ = 0.
                for p in p_vals:
                    for q in q_vals:
                        alpha = (1 - (I[p, q] - I[i, j]))
                        x = I[p, q] * np.cos((np.pi * 2) *(alpha - S[i, j]))
                        y = 1 / (lamb + np.exp(-self.mu * (x - S[i, j])))
                        sum_ += y
                        alpha_+=alpha
                #print('Lambda: ',lamb, 'x: ', x, 'S: ', S[i,j])
                f[i, j] = sum_
                Alpha[i, j] = alpha_

        return f, Alpha


    def select_image(self, mae_list, image_list):
        """
        Selects the output of the image transformation after multiple iterations.

        The selection is based on detecting peaks in the Mean Absolute Differences (MAD).

        Args:
            mae_list (list): List of Mean Absolute Differences between iterations.
            image_list (list): List of image transformations.

        Returns:
            np.array: Final selected image.

            int: Iteration index at which the best image was selected.
        """
        # We select the image that has a peak in the mae
        mae_list = mae_list[1:]
        if len(mae_list)==1:
            return image_list[1], 1
        peaks, _ = find_peaks(mae_list, height=0)
        if len(peaks)==0:
            selected_peak = np.argmin(np.abs(np.diff(mae_list))/mae_list[1:]) + 1
        else:
            # If there are more than 1 peaks, we select the mean value
            selected_peak = int(np.mean(peaks))
            # For longer iterations, we add 1 to the output
            if selected_peak>5:
                selected_peak+= 1
        # Return the final selected image
        selected_peak = max(selected_peak,1)
        # If the images converge to a constant color, choose intermediate steps
        density, _ = np.histogram(image_list[selected_peak].flatten()/255, bins=20, density=False)
        density = density/np.sum(density)
        if np.max(density)>0.8:
            for i in range(len(image_list)-1, 0, -1):
                density, _ = np.histogram(image_list[i].flatten()/255, bins=20, density=False)
                density = density/np.sum(density)
                if np.max(density)<0.75:
                    selected_peak = i
                    return image_list[selected_peak], selected_peak
            selected_peak = 1
        return image_list[selected_peak], selected_peak


    def process(self, im, save=True, image_path='./filtered_image.png'):
        """
        Applies the quantum-inspired transformation until convergence.

        Args:
            im (np.array): Input image.
            save (bool, optional): If True, saves the final image. Defaults to True.
            image_path (str, optional): Path to save the final image. Defaults to './filtered_image.png'.

        Returns:
            np.array: Final processed image.

            list: List of Mean Absolute Differences between iterations.

            list: List of image transformations.
        """
        
        wt, ht = im.shape[1], im.shape[0]  # Dimensions of the matrix    
        self.info['num_circuits'] = wt*ht
        self.info['nqbits'] = 3
        self.info['depth'] = 4
        self.info['gates'] = {'RX':2, 'CCX': 1, 'H': 2}
        self.info['num_nonlocal_gates'] = 1
        # Determine cluster values w
        min_val, max_val = np.percentile(im.flatten()/255, self.percentile), 1
        dw = (max_val - min_val)/(self.L-3)
        w = [0] + list(np.arange(min_val, max_val+dw, dw)) 
        
        transformed_image = im.copy().astype(float) / 255  # Initial image 
        #transformed_image *= (np.pi / 2) # Normalize initial image to [0, pi/2]

        w1 = np.zeros((ht, wt))
        mae_list = [0]
        image_list = [im]
        # Evolve Quantum inspired image until convergence
        for t in range(self.max_iter):  # Limit iterations 
            if t==0:
                cluster=w
            else:
                cluster = [0 ,0.16, 0.32, 0.48 ,0.64, 0.80, 0.96, 1]

            intermediate_matrix, _, = self.transformation(transformed_image, cluster)  
            transformed_image, w2 = self.transformation(intermediate_matrix, cluster)
            stop, mae = self.check_convergence(w2, w1)
            mae_list.append(mae)
            image_list.append((transformed_image * 255).astype(np.uint8))
            if stop:# Check for convergence
                break
            w1 = w2
        print('Number of iterations: ', t+1)
        out, selected_image=  self.select_image(mae_list, image_list)
        print('Selected image: ', selected_image)
        if save:
            Image.fromarray(out).save(image_path)

        return out, mae_list, image_list




from numpy.lib.stride_tricks import as_strided as numpy_as_strided
from torch import as_strided as torch_as_strided


def _roll_shape_and_strides(
        input_shape, input_strides, window_shape, dx=1, dy=1, dz=None):
    '''
    Rolling 3D window for array.

    Args:
        input_shape (np.array): input array shape
        input_strides (np.array): input array strides
        window_shape (np.array): rolling 2D window array shape
        dx (int): horizontal step, abscissa, number of columns
        dy (int): vertical step, ordinate, number of rows
        dz (int): transverse step, applicate, number of layers. Only used with
        3D window

    Returns:
        tuple: Shape of the rolling window
        np.array: Stride of the rolling window
    '''
    if dz is not None:
        shape = input_shape[:-3] + (
            (input_shape[-3] - window_shape[-3]) // dz + 1,
        )
        strides = input_strides[:-3] + (input_strides[-3] * dz,)
    else:
        shape = input_shape[:-2]
        strides = input_strides[:-2]

    shape += \
        ((input_shape[-2] - window_shape[-2]) // dy + 1,) + \
        ((input_shape[-1] - window_shape[-1]) // dx + 1,) + \
        window_shape  # multidimensional "sausage" with 3D cross-section
    strides += \
        (input_strides[-2] * dy,) + (input_strides[-1] * dx,)

    if dz is not None:
        strides += input_strides[-3:]
    else:
        strides += input_strides[-2:]

    return shape, strides


def roll_numpy(input_array, window_array, dx=1, dy=1, dz=None):
    '''
    Rolling 3D window for numpy array. This function is only used with
    Qiskit backends.

    Args:
        input_array (np.array): input array
        window_array (np.array): rolling 2D window array
        dx (int): horizontal step, abscissa, number of columns
        dy (int): vertical step, ordinate, number of rows
        dz (int): transverse step, applicate, number of layers. Only used with
        3D window

    Returns:
        np.array: Stride of the rolling window
        tuple: Shape of the rolling window       
    '''
    shape, strides = _roll_shape_and_strides(
        input_array.shape, input_array.strides, window_array.shape,
        dx, dy, dz)
    return numpy_as_strided(input_array, shape=shape, strides=strides), shape


def roll_torch(input_array, window_array, dx=1, dy=1, dz=None):
    '''
    Rolling 3D window for pytorch tensor. This function is only used with
    Pytorch backends.

    Args:
        input_array (tensor): input array, shape (n_samples, N,N,N)
        window_array (tensor): rolling 3D window array, shape (n,n,n)
        dx (int): horizontal step, abscissa, number of columns
        dy (int): vertical step, ordinate, number of rows
        dz (int): transverse step, applicate, number of layers
        
    Returns:
        tensor: Stride of the rolling window
        tuple: Shape of the rolling window 
    '''
    shape, strides = _roll_shape_and_strides(
        input_array.shape, input_array.stride(), window_array.shape,
        dx, dy, dz)
    return torch_as_strided(input_array, shape, strides), shape



import numpy as np
import networkx as nx
from scipy.spatial.distance import canberra
from scipy.special import rel_entr
from sklearn.preprocessing import MinMaxScaler


def euclidean_distance(p1, p2):
    ''' 
    Euclidean distance.

    Args:
      p1 (float): pixel value 1
      p2 (float): pixel value 2

    Returns:
      float: Euclidean distance between p1 and p2
    '''
    return np.sqrt(np.sum((p1 - p2) ** 2))

def minkowski_distance(p1, p2, r):
    ''' 
    Minkowski distance.

    Args:
      p1 (float): pixel value 1
      p2 (float): pixel value 2

    Returns:
      float: Minkowski distance between p1 and p2
    '''
    return np.power(np.sum(np.power(np.abs(p1 - p2), r)), 1/r)

def chebyshev_distance(p1, p2):
    ''' 
    Chebyshev distance.

    Args:
      p1 (float): pixel value 1
      p2 (float): pixel value 2

    Returns:
      float: Chebyshev distance between p1 and p2
    '''
    return np.max(np.abs(p1 - p2))

def jensen_shannon_divergence(X, Y):
    ''' 
    Jensen-Shannon Divergence.

    Args:
      p1 (float): pixel value 1
      p2 (float): pixel value 2

    Returns:
      float: Jensen-Shannon Divergence between p1 and p2
    '''
    M = 0.5 * (X + Y)
    return np.sqrt(np.abs(0.5 * (rel_entr(X, M).sum() + rel_entr(Y, M).sum())))

def gaussian_similarity(a, b, sigma):
  ''' 
  Gaussian similarity.

  Args:
    p1 (float): pixel value 1
    p2 (float): pixel value 2

  Returns:
    float: Gaussian similarity between p1 and p2
  '''
  gaussian_similairity_score = np.exp(-((a - b)**2) / (2 * sigma**2))
  return gaussian_similairity_score

# Canberra Distance
def canberra_distance(X, Y):
    ''' 
    Canberra distance.

    Args:
      p1 (float): pixel value 1
      p2 (float): pixel value 2

    Returns:
      float: Canberra distance between p1 and p2
    '''
    return canberra(X, Y)

def apply_similarity(type,image, x1, y1, x2, y2, sigma=None):
  ''' 
  Apply chosen similarity to image.

  Args:
    type (str): Similarity type name
    image (np.array): input image
    x1 (int): x coordinates pixel 1
    y1 (int): y coordinates pixel 1
    x2 (int): x coordinates pixel 2
    y2 (int): y coordinates pixel 2

  Returns:
    weight (float): Edge weight
  '''
  if type=='Gaussian':
    weight = -1*(1-gaussian_similarity(image[x1,y1],image[x2,y2], sigma))
  elif type=='Euclidean':
    weight = (1-euclidean_distance(image[x1,y1],image[x2,y2]))
  elif type=='Minkowski':
    weight = (1-minkowski_distance(image[x1,y1].reshape(1),image[x2,y2].reshape(1),3))
  elif type=='Chebyshev':
    weight = (1-chebyshev_distance(image[x1,y1].reshape(1),image[x2,y2].reshape(1)))
  elif type=='Jensen':
    weight = (1-jensen_shannon_divergence(np.abs(image[x1,y1]).reshape(1),np.abs(image[x2,y2]).reshape(1)))
  elif type=='Canberra':
    weight = (1-canberra_distance(image[x1,y1].reshape(1),image[x2,y2].reshape(1)))
  return weight



def image_to_grid_graph(gray_img, type = 'Gaussian', sigma = None):
  """
  Convert a grayscale image to a grid graph with Gaussian similarity as edge weights.

  Args:
    gray_img (numpy.ndarray): Grayscale image.
    type (str): Similarity type name
    sigma (float): Parameter for Gaussian similarity.

  Returns:
    list: List of edges with weights for the graph.
  """
  h, w = gray_img.shape  
  if sigma is None and type=='Gaussian': # Calculate std as the data std unless stated otherwise
    data = gray_img.flatten()
    sigma = np.std(data)
  nodes = np.zeros((h*w, 1))
  edges = []
  nx_elist = []
  # Calculate similarity between neighboring pixels
  for i in range(h*w):
    x, y = i//w, i%w
    nodes[i] = gray_img[x,y]
    if x > 0:
      j = (x-1)*w + y
      weight = apply_similarity(type,gray_img, x,y, x-1, y,sigma)
      edges.append((i, j, weight))
      nx_elist.append(((x,y),(x-1,y),np.round(weight,2)))
    if y > 0:
      j = x*w + y-1
      weight = apply_similarity(type,gray_img, x,y, x, y-1,sigma)
      edges.append((i, j, weight))
      nx_elist.append(((x,y),(x,y-1),weight))
  w_list = [e[2] for e in edges]
  # Normalize weights to the [-1,1] domain
  scaler = MinMaxScaler((-1,1))
  w_list_scaled = scaler.fit_transform(np.array(w_list).reshape(-1,1)).flatten()
  normalized_nx_elist = []
  for i in range(len(nx_elist)):
    normalized_nx_elist.append((nx_elist[i][0], nx_elist[i][1], w_list_scaled[i]))
  return normalized_nx_elist


def create_graph(data_small, type = 'Gaussian', sd_prop = 0.3):
  '''
  Function to create the Graph encoding the image, where the edges are calculated using the Gaussian similarity between pixels.
  Here we perform hyperparameter optimization for the stadard deviation, ensuring that  the proportion of the two tails of the weights distribution is around 0.17. 

  Args:
    data_small (np.array): input data
    type (str): Similarity type name

  Returns:
    G (networkx.graph): Graph encoding the image to be segmented.
  '''
  x = data_small.flatten()
  if type=='Gaussian':
    normalized_nx_elist = image_to_grid_graph(data_small, sigma=sd_prop*np.std(x)) 
    G = nx.grid_2d_graph(data_small.shape[0], data_small.shape[1])
    G.add_weighted_edges_from(normalized_nx_elist)
  else: # For other similarity measures there is no hyperparameter search
    normalized_nx_elist = image_to_grid_graph(data_small, type)
    G = nx.grid_2d_graph(data_small.shape[0], data_small.shape[1])
    G.add_weighted_edges_from(normalized_nx_elist)
    std_p =1
  return G



import numpy as np
import torch

from itertools import product, permutations
from math import sqrt
from qiskit import QuantumCircuit
from qiskit_aer import AerSimulator
from qiskit_ibm_runtime import Session, SamplerV2 as Sampler
from qiskit.transpiler.preset_passmanagers import generate_preset_pass_manager
from tqdm import tqdm
from time import time

from .utils import roll_numpy, roll_torch

class EdgeDetectorBase:
    '''
    Base class for quantum-inspired edge detection.

    This class implements both Pytorch-based and Qiskit-based quantum Hadamard 
    edge detection algorithms for images and 3D volumes.

    Attributes:
        n_dimensions (int): Number of dimensions (2D or 3D).
        size (int): Size of the (NxN) blocks used for partitioning the image.
        backend (str): Computational backend ('torch' for PyTorch, 'aer_simulator' for Qiskit).
        shots (int): Number of shots for Qiskit experiments.
        total_qb (int): Total number of qubits used in the quantum circuit.
        anc_qb (int): Number of ancilla qubits.
        data_qb (int): Number of qubits required for data encoding.
        device (torch.device, optional): PyTorch device ('cpu' or 'cuda'), only for `torch` backend.

    '''
    def __init__(self, n_dimensions: int, size: int, backend: str, shots: int):
        """
        Initializes the EdgeDetectorBase class.

        Args:
            n_dimensions (int): Number of dimensions (2D or 3D).
            size (int): Size of the (NxN) blocks for partitioning the image.
            backend (str): 'torch' for PyTorch or 'aer_simulator' for Qiskit.
            shots (int): Number of shots for Qiskit experiments.

        Raises:
            ValueError: If an invalid backend is provided.
        """

        self.n_dimensions = n_dimensions

        # 1. Circuit parameters: data qubits  ancilla qubits
        self.size = size
        self.shape = (size,) * n_dimensions
        self.data_qb = int(np.log2(self.size**n_dimensions))
        self.anc_qb = 1
        self.total_qb = self.data_qb + self.anc_qb
        self.backend = backend

        # Running in Pytorch
        if backend == 'torch':
            # 1. set CUDA for PyTorch
            use_cuda = torch.cuda.is_available()
            if use_cuda:
                self.device = torch.device("cuda:0")
                torch.cuda.set_device(int("cuda:0".split(':')[1]))
            else:
                self.device = torch.device("cpu")

            # 2. Quantum operations: Hadamard and rolling Identity
            hadamard = 1/sqrt(2)*torch.tensor(
                [[1, 1], [1, -1]], dtype=torch.float32
            ).to(self.device)

            self.hadamard_large = torch.kron(
                torch.eye(2**self.data_qb), hadamard
            ).reshape(
                1, 2**self.total_qb, 2**self.total_qb
            ).to(self.device)

            # Initialize the amplitude permutation unitary
            self.D2n = torch.roll(
                torch.eye(2**self.total_qb, dtype=torch.float32), 1, 1
            ).reshape(
                1, 2**self.total_qb, 2**self.total_qb
            ).to(self.device)

        # Running in a Qiskit environment
        elif backend=='aer_simulator':
            self.backend = AerSimulator()
        elif type(backend) == str:
            raise ValueError('The only valid backend names are "torch" and "aer_simulator". You can also provide a Qiskit Backend directly.')
        self.shots = shots
        # Quantum operations: rolling Identity
        self.D2n = np.roll(np.identity(2**self.total_qb), 1, axis=1)
            
    def run_box(self, box, tol=1e-3):
        """
        Runs the edge detection algorithm for a single box of the image.

        This function executes a quantum Hadamard-based edge detection 
        algorithm for a given image segment.

        Args:
            box (np.array): Input image segment.
            tol (float, optional): Threshold for detecting edges. Defaults to 1e-3.

        Returns:
            np.array: Edge-detected version of the input box.
        """
        # Create quantum circuit
        qc = QuantumCircuit(self.total_qb)

        # Data encoding: amplitude encoding
        box = box.flatten()
        initial_state = np.zeros(2**self.data_qb)
        initial_state[:box.shape[0]] = box
        initial_state = initial_state/np.linalg.norm(initial_state)
        qc.initialize(initial_state, range(1, self.total_qb))

        # Apply quantum operations
        qc.h(0)
        qc.unitary(self.D2n, range(self.total_qb))
        qc.h(0)

        # Measure
        qc.measure_all()

        # Run quantum circuit
        is_simulator = self.backend.configuration().simulator # Check if we're using a simulator or quantum hardware
        with Session(backend=self.backend) as session: # Create a session with such backend
            sampler = Sampler(mode=session)
            sampler.options.default_shots = self.shots

            # If running on real hardware (not a simulator), apply error mitigation
            if not is_simulator:
                sampler.options.dynamical_decoupling.enable = True
                sampler.options.dynamical_decoupling.sequence_type = "XY4"
                sampler.options.twirling.enable_gates = True
                sampler.options.twirling.num_randomizations = "auto"

            # Transpile circuit for noisy basis gates
            passmanager = generate_preset_pass_manager(
                optimization_level=2, backend=self.backend
            )
            qc_noise = passmanager.run(qc)

            pub = (qc_noise,) # Prepare input
            job = sampler.run([pub]) 
            counts = job.result()[0].data.meas.get_counts()

        # Get statevector from counts
        # Calculate binary string of basis qubits
        # Let's calculate the ordered statevector
        new_counts = [
            counts.get(
                ''.join(map(str, qbits)),
                0
            )  # Generate string link e.g. '000', '001', ...
            for qbits in product([0, 1], repeat=self.total_qb)
        ]
        statevector = np.array(new_counts)/np.sum(new_counts)

        # Get odd values from state
        final_state = statevector[range(1, 2**(self.data_qb + 1), 2)]

        # Select values larger than threshold
        edge_scan = np.zeros(final_state.shape)
        edge_scan[np.abs(final_state) > tol] = 1

        # Revert to box shape
        result = edge_scan[:self.size**self.n_dimensions].reshape(self.shape)
        return result

    def run_image_qiskit(self, data, tol=1e-3, reduce=True, verbose=False):
        """
        Runs the edge detection algorithm for the entire image using Qiskit.

        Args:
            data (np.array): Input image.
            tol (float, optional): Threshold for detecting edges. Defaults to 1e-3.
            reduce (bool, optional): If True, reduces the image dimensions by half. Defaults to True.
            verbose (bool, optional): If True, displays progress using tqdm. Defaults to False.

        Returns:
            np.array: Edge-detected image.
        """
        # 1. Get view of image
        # We split it into bits of (4x4x4)
        windows, shape_aux = roll_numpy(
            data, np.zeros((self.size,) * self.n_dimensions),
            dx=self.size, dy=self.size,
            dz=self.size if self.n_dimensions == 3 else None
        )
        windows = windows.reshape((-1,) + (self.size,) * self.n_dimensions)

        # 2. Run every box thorugh the QC
        def get_box_result(box):
            if np.sum(box) > tol:
                return self.run_box(box, tol)
            else:
                return np.zeros(self.shape)

        if verbose:
            results = [
                get_box_result(windows[i])
                for i in tqdm(range(windows.shape[0]))
            ]
        else:
            results = [
                get_box_result(windows[i])
                for i in range(windows.shape[0])
            ]

        # 3. Reshape image to original shape
        results = np.array(results).reshape(shape_aux)

        transpose_idxs_map = {
            4: [0, 2, 1, 3],
            5: [0, 1, 3, 2, 4],
            6: [0, 3, 1, 4, 2, 5],
            7: [0, 1, 4, 2, 5, 3, 6],
            8: [0, 1, 2, 5, 3, 6, 4, 7],
        }
        transpose_idxs = transpose_idxs_map[len(results.shape)]

        reverted_image = results.transpose(transpose_idxs).reshape(data.shape)

        if not reduce:
            return reverted_image

        # 4. Reduce image size
        size_x = data.shape[-1]
        reverted_image_small = reverted_image[
            :, :, :, :, range(1, size_x, 2)
        ][
            :, :, :, range(1, size_x, 2), :
        ]

        if self.n_dimensions == 3:
            reverted_image_small = \
                reverted_image_small[:, :, range(1, size_x, 2), :, :]

        return reverted_image_small

    def run_image_torch(self, data, tol=1e-3, reduce=True, verbose=False):
        """
        Runs edge detection for the entire image using PyTorch.

        Args:
            data (torch.Tensor): Input image.
            tol (float, optional): Threshold for detecting edges. Defaults to 1e-3.
            reduce (bool, optional): If True, reduces the image dimensions by half. Defaults to True.
            verbose (bool, optional): If True, displays progress using tqdm. Defaults to False.

        Returns:
            torch.Tensor: Edge-detected image.
        """
        if verbose:
            start_time = time()

        if self.n_dimensions == 2:
            axis = (1, 2)
            reshape_idxs = (-1, 1, 1)
            normalized_reshape_idxs = (-1, self.size**self.n_dimensions)
            windows_reshape_idxs = (-1, self.size, self.size)
        else:
            axis = (2, 3, 4)
            samples = data.shape[0]
            reshape_idxs = (samples, -1, 1, 1, 1)
            normalized_reshape_idxs = \
                (samples, -1, self.size**self.n_dimensions)
            windows_reshape_idxs = \
                (samples, -1, self.size, self.size, self.size)

        # 1. Get view of image
        # We split it into bits of (4x4x4)
        windows, shape_aux = roll_torch(
            data, torch.zeros((self.size,) * self.n_dimensions),
            dx=self.size, dy=self.size,
            dz=self.size if self.n_dimensions == 3 else None
        )
        windows = windows.reshape(windows_reshape_idxs).to(self.device)

        # 2. Data encoding: amplitude encoding
        norm_const = windows.square().sum(axis=axis) \
            .sqrt().reshape(*reshape_idxs)
        norm_const[torch.abs(norm_const) < 1e-16] = 1
        normalized_image = windows/norm_const

        # 3. Flatten the last 2 dimesions to obtain an initial state
        normalized_flatten = normalized_image.reshape(normalized_reshape_idxs)

        # 4. Kronecker product to increase the last dimension 1 qubit
        normalized_larger = torch.kron(
            torch.tensor(normalized_flatten, dtype=torch.float32),
            torch.tensor([1, 0]).to(self.device)
        )
        self.D2n = torch.tensor(self.D2n, dtype = torch.float32).to(self.device)
        # 5. Apply quantum operations
        final_state_h = (
            normalized_larger @ self.hadamard_large @ self.D2n
            @ self.hadamard_large
        )[:, :, range(1, 2**(self.data_qb+1), 2)]

        # 6. Select values larger than threshold
        edge_scan_h = torch.zeros(final_state_h.shape).to(self.device)
        edge_scan_h[torch.abs(final_state_h) > tol] = 1

        # 7. Reshape image to original shape
        permutations_idxs = {
            2: [0, 1, 3, 2, 4],
            3: [0, 1, 2, 5, 3, 6, 4, 7],
        }
        reverted_image = edge_scan_h \
            .reshape(shape_aux) \
            .permute(permutations_idxs[self.n_dimensions]) \
            .reshape(data.shape)
        if self.n_dimensions == 3:
            reverted_image[:, :, :, :, 0] = 0

        # 8. Reduce image size
        if reduce:
            size_x = data.shape[-1]
            reverted_image = reverted_image[
                :, :, :, :, range(1, size_x, 2)
            ][
                :, :, :, range(1, size_x, 2), :
            ]

            if self.n_dimensions == 3:
                reverted_image = \
                    reverted_image[:, :, range(1, size_x, 2), :, :]

        if verbose:
            print('Total execution time: ', time() - start_time)

        return reverted_image


###############################################################################
# QUANTUM EDGE DETECTION ALGORITHM FOR 2D IMAGES
###############################################################################

class EdgeDetector2D(EdgeDetectorBase):
    """
    Quantum Hadamard Edge Detection algorithm for 2D images.

    This class implements both Qiskit-based and PyTorch-based quantum edge detection 
    algorithms for two-dimensional images.

    Attributes:
        size (int): Size of the (NxN) blocks used for partitioning the image.
        backend (str): Computational backend ('torch' for PyTorch, 'aer_simulator' for Qiskit).
        shots (int): Number of shots for Qiskit experiments.
    """

    def __init__(self, size, backend='aer_simulator', shots=1000):
        """
        Initializes the 2D Quantum Hadamard Edge Detection algorithm.

        Args:
            size (int): Size of the (NxN) blocks for partitioning the image.
            backend (str, optional): 'torch' for PyTorch or 'aer_simulator' for Qiskit. Defaults to 'aer_simulator'.
            shots (int, optional): Number of shots for Qiskit experiments. Defaults to 1000.
        """

        super().__init__(
            n_dimensions=2, size=size, backend=backend, shots=shots)

    def run(self, data, tol=1e-3, reduce=True, verbose=False):
        """
        Runs the edge detection algorithm with either Pytorch or Qiskit code

        Args:
            data (tensor/np.array): input data
            tol (tensor/np.array): Tolerance to be considered and edge
            reduce (bool): reduce the dimension by half at the end
            verbose (bool): If true, tqdm is used to show the evolution.
            
        Returns:
            np.array/tensor: Edge-detected image
        """
        if self.backend == 'torch':
            return self.run_image_torch(data, tol, reduce, verbose)
        return self.run_image_qiskit(data, tol, reduce, verbose)


###############################################################################
# QUANTUM EDGE DETECTION ALGORITHM FOR 3D VOLUMES
###############################################################################

class EdgeDetector3D(EdgeDetectorBase):
    """
    Quantum Hadamard Edge Detection algorithm for 3D volumes.

    This class implements both Qiskit-based and PyTorch-based quantum edge detection 
    algorithms for three-dimensional images.

    Attributes:
        size (int): Size of the (NxN) blocks used for partitioning the image.
        backend (str): Computational backend ('torch' for PyTorch, 'aer_simulator' for Qiskit).
        shots (int): Number of shots for Qiskit experiments.
    """

    def __init__(self, size, backend='aer_simulator', shots=100):
        """
        Initializes the 3D Quantum Hadamard Edge Detection algorithm.

        Args:
            size (int): Size of the (NxN) blocks for partitioning the image.
            backend (str, optional): 'torch' for PyTorch or 'aer_simulator' for Qiskit. Defaults to 'aer_simulator'.
            shots (int, optional): Number of shots for Qiskit experiments. Defaults to 100.
        """

        super().__init__(
            n_dimensions=3, size=size, backend=backend, shots=shots)

    def run(self, data, num_filters=6, tol=1e-3, reduce=True, verbose=False):
        '''
        Run edge detection for different rotations of the image.

        Args:
            data (np.array): input data (samples, features, dim,dim,dim)
            num_filters (int): Number of rotations (permutations of the last 3
            dimension) to output
            tol (float):Tolerance to consider an edge
            reduce (bool): reduce the size of the image at the end
            verbose (bool): If true, tqdm is used to show the evolution.
        Returns:
            np.array/tensor: Edge-detected image
        '''
        if reduce:
            fin_shape = int(data.shape[-1] / 2)
        else:
            fin_shape = int(data.shape[-1])
        perm = list(permutations([2, 3, 4]))
        if self.backend == 'torch':
            data_out = torch.zeros((
                data.shape[0], data.shape[1]*num_filters,
                fin_shape, fin_shape, fin_shape))
        else:
            data_out = np.zeros((
                data.shape[0], data.shape[1]*num_filters,
                fin_shape, fin_shape, fin_shape))
        for i in range(num_filters):
            p = [0, 1] + list(perm[i])
            if self.backend == 'torch':
                data_out[:, data.shape[1]*i:data.shape[1]*(i+1), :, :, :] = \
                    self.run_image_torch(
                        data.permute(p), tol, reduce, verbose).permute(p)
            else:
                data_out[:, data.shape[1]*i:data.shape[1]*(i+1), :, :, :] = \
                    self.run_image_qiskit(
                        data.transpose(p), tol, reduce, verbose).transpose(p)
        return data_out



from scipy.optimize import minimize_scalar
import tensorly as tl
import numpy as np


def estimate_ranks(layer):
    """ 
    Estimate the optimal ranks for Tucker decomposition of a tensor using Variational Bayesian Matrix Factorization (VBMF).
    This function unfolds the input tensor along two modes (rows and columns) and applies VBMF to estimate 
    the optimal ranks of the matrices corresponding to each mode. The estimated ranks are then returned 
    as a list.

    Args:
        layer (torch.nn.Module): A neural network layer whose weights are represented as a tensor.
        
    Returns:
        list: A list containing the estimated ranks for the decomposition along the first and second modes.
    """

    weights = layer.weight.data.numpy()
    if weights.shape[0]==1 and weights.shape[1]>1:
        return [1, weights.shape[1]//2]
    if weights.shape[0]>1 and weights.shape[1]==1:
        return [ weights.shape[0]//2, 1]
    if weights.shape[0]==1 and weights.shape[1]==1:
        return [1, 1]
    unfold_0 = tl.base.unfold(weights, 0) 
    unfold_1 = tl.base.unfold(weights, 1)
    _, diag_0, _, _ = EVBMF(unfold_0)
    _, diag_1, _, _ = EVBMF(unfold_1)
    ranks = [diag_0.shape[0], diag_1.shape[1]]
    return ranks



def VBMF(Y, cacb, sigma2=None, H=None):
    """Implementation of the analytical solution to Variational Bayes Matrix Factorization.

    This function can be used to calculate the analytical solution to VBMF. 
    This is based on the paper and MatLab code by Nakajima et al.:
    "Global analytic solution of fully-observed variational Bayesian matrix factorization."

    Args:
        Y (numpy-array): Input matrix that is to be factorized. Y has shape (L,M), where L<=M.   
        cacb (int): Product of the prior variances of the matrices that factorize the input.
        sigma2 (int or None): Variance of the noise on Y.
        
        H (int or None): Maximum rank of the factorized matrices.
        
    Returns:
        U (numpy-array): Left-singular vectors. 
        S (numpy-array): Diagonal matrix of singular values.  
        V (numpy-array): Right-singular vectors.  
        (post dictionary): Dictionary containing the computed posterior values.
        
    References:
        [1] Nakajima, Shinichi, et al. "Global analytic solution of fully-observed variational Bayesian matrix factorization." Journal of Machine Learning Research 14.Jan (2013): 1-37.
        [2] Nakajima, Shinichi, et al. "Perfect dimensionality recovery by variational Bayesian PCA." Advances in Neural Information Processing Systems. 2012.
    """    
    
    L,M = Y.shape # has to be L<=M

    if H is None:
        H = L
    
    # SVD of the input matrix, max rank of H
    U,s,V = np.linalg.svd(Y)
    U = U[:,:H]
    s = s[:H]
    V = V[:H].T 

    # Calculate residual
    residual = 0.
    if H<L:
        residual = np.sum(np.sum(Y**2)-np.sum(s**2))

    # Estimation of the variance when sigma2 is unspecified
    if sigma2 is None: 
        upper_bound = (np.sum(s**2)+ residual)/(L+M)

        if L==H: 
            lower_bound = s[-1]**2/M
        else:
            lower_bound = residual/((L-H)*M)

        sigma2_opt = minimize_scalar(VBsigma2, args=(L,M,cacb,s,residual), bounds=[lower_bound, upper_bound], method='bounded')
        sigma2 = sigma2_opt.x
        print("Estimated sigma2: ", sigma2)

    # Threshold gamma term
    # Formula above (21) from [1]
    thresh_term = (L+M + sigma2/cacb**2)/2 
    threshold = np.sqrt( sigma2 * (thresh_term + np.sqrt(thresh_term**2 - L*M) ))
              
    # Number of singular values where gamma>threshold
    pos = np.sum(s>threshold)

    # Formula (10) from [2]
    d = np.multiply(s[:pos], 
                    1 - np.multiply(sigma2/(2*s[:pos]**2),
                                    L+M+np.sqrt( (M-L)**2 + 4*s[:pos]**2/cacb**2 )))

    # Computation of the posterior
    post = {}
    zeta = sigma2/(2*L*M) * (L+M+sigma2/cacb**2 - np.sqrt((L+M+sigma2/cacb**2)**2 - 4*L*M))
    post['ma'] = np.zeros(H) 
    post['mb'] = np.zeros(H)
    post['sa2'] = cacb * (1-L*zeta/sigma2) * np.ones(H)
    post['sb2'] = cacb * (1-M*zeta/sigma2) * np.ones(H)  

    delta = cacb/sigma2 * (s[:pos]-d- L*sigma2/s[:pos])
    post['ma'][:pos] = np.sqrt(np.multiply(d, delta))
    post['mb'][:pos] = np.sqrt(np.divide(d, delta))
    post['sa2'][:pos] = np.divide(sigma2*delta, s[:pos])
    post['sb2'][:pos] = np.divide(sigma2, np.multiply(delta, s[:pos]))
    post['sigma2'] = sigma2
    post['F'] = 0.5*(L*M*np.log(2*np.pi*sigma2) + (residual+np.sum(s**2))/sigma2 - (L+M)*H
               + np.sum(M*np.log(cacb/post['sa2']) + L*np.log(cacb/post['sb2'])
                        + (post['ma']**2 + M*post['sa2'])/cacb + (post['mb']**2 + L*post['sb2'])/cacb
                        + (-2 * np.multiply(np.multiply(post['ma'], post['mb']), s)
                           + np.multiply(post['ma']**2 + M*post['sa2'],post['mb']**2 + L*post['sb2']))/sigma2))

    return U[:,:pos], np.diag(d), V[:,:pos], post


def VBsigma2(sigma2, L, M, cacb, s, residual):
    """
    Computes the free energy of the Variational Bayesian Matrix Factorization (VBMF) model 
    to optimize the noise variance parameter.

    Args:
        sigma2 (float): Variance of the noise in the model.
        L (int): Number of rows in the input matrix.
        M (int): Number of columns in the input matrix.
        cacb (float): Multiplicative factor derived from prior hyperparameters.
        s (numpy.ndarray): Singular values of the matrix to factorize.
        residual (float): Residual sum of squares from truncated singular values.

    Returns:
        float: Free energy of the VBMF model for the given noise variance.
    """
    H = len(s)

    thresh_term = (L + M + sigma2 / cacb ** 2) / 2
    threshold = np.sqrt(sigma2 * (thresh_term + np.sqrt(thresh_term ** 2 - L * M)))
    pos = np.sum(s > threshold)
    
    d = np.multiply(s[:pos], 
                    1 - np.multiply(sigma2 / (2 * s[:pos] ** 2),
                                    L + M + np.sqrt((M - L) ** 2 + 4 * s[:pos] ** 2 / cacb ** 2)))

    zeta = sigma2 / (2 * L * M) * (L + M + sigma2 / cacb ** 2 - np.sqrt((L + M + sigma2 / cacb ** 2) ** 2 - 4 * L * M))
    post_ma = np.zeros(H) 
    post_mb = np.zeros(H)
    post_sa2 = cacb * (1 - L * zeta / sigma2) * np.ones(H)
    post_sb2 = cacb * (1 - M * zeta / sigma2) * np.ones(H)  

    delta = cacb / sigma2 * (s[:pos] - d - L * sigma2 / s[:pos])
    post_ma[:pos] = np.sqrt(np.multiply(d, delta))
    post_mb[:pos] = np.sqrt(np.divide(d, delta))
    post_sa2[:pos] = np.divide(sigma2 * delta, s[:pos])
    post_sb2[:pos] = np.divide(sigma2, np.multiply(delta, s[:pos]))

    F = 0.5 * (L * M * np.log(2 * np.pi * sigma2) + (residual + np.sum(s ** 2)) / sigma2 - (L + M) * H
               + np.sum(M * np.log(cacb / post_sa2) + L * np.log(cacb / post_sb2)
                        + (post_ma ** 2 + M * post_sa2) / cacb + (post_mb ** 2 + L * post_sb2) / cacb
                        + (-2 * np.multiply(np.multiply(post_ma, post_mb), s)
                           + np.multiply(post_ma ** 2 + M * post_sa2, post_mb ** 2 + L * post_sb2)) / sigma2))
    return F


def EVBMF(Y, sigma2=None, H=None):
    """
    Analytical implementation of Empirical Variational Bayesian Matrix Factorization (VBMF).

    Args:
        Y (numpy.ndarray): Input matrix to factorize with shape (L, M), where L <= M.
        sigma2 (float, optional): Noise variance. If None, it is estimated.
        H (int, optional): Maximum rank of the factorized matrices. If None, set to L.

    Returns:
        tuple: 
            - U (numpy.ndarray): Left singular vectors of the factorized matrix.
            - S (numpy.ndarray): Diagonal matrix of singular values.
            - V (numpy.ndarray): Right singular vectors of the factorized matrix.
            - post (dict): Dictionary containing posterior values including 'sigma2' and 'F'.

    References:
        - Nakajima, Shinichi, et al. "Global analytic solution of fully-observed variational 
          Bayesian matrix factorization." Journal of Machine Learning Research 14.Jan (2013).
    """
    L, M = Y.shape  # has to be L <= M

    if H is None:
        H = L

    alpha = L / M
    tauubar = 2.5129 * np.sqrt(alpha)
    
    # SVD of the input matrix, max rank of H
    U, s, V = np.linalg.svd(Y)
    U = U[:, :H]
    s = s[:H]
    V = V[:H].T 

    # Calculate residual
    residual = 0.
    if H < L:
        residual = np.sum(np.sum(Y ** 2) - np.sum(s ** 2))

    # Estimation of the variance when sigma2 is unspecified
    if sigma2 is None: 
        xubar = (1 + tauubar) * (1 + alpha / tauubar)
        eH_ub = int(np.min([np.ceil(L / (1 + alpha)) - 1, H])) - 1
        upper_bound = (np.sum(s ** 2) + residual) / (L * M)
        lower_bound = np.max([s[eH_ub + 1] ** 2 / (M * xubar), np.mean(s[eH_ub + 1:] ** 2) / M])

        scale = 1.  # / lower_bound
        s = s * np.sqrt(scale)
        residual = residual * scale
        lower_bound = lower_bound * scale
        upper_bound = upper_bound * scale

        sigma2_opt = minimize_scalar(EVBsigma2, args=(L, M, s, residual, xubar), bounds=[lower_bound, upper_bound], method='Bounded')
        sigma2 = sigma2_opt.x *0.5

    # Threshold gamma term
    threshold = np.sqrt(M * sigma2 * (1 + tauubar) * (1 + alpha / tauubar))
    pos = np.sum(s > threshold)

    # Formula (15) from [2]
    d = np.multiply(s[:pos] / 2, 1 - np.divide((L + M) * sigma2, s[:pos] ** 2) + np.sqrt((1 - np.divide((L + M) * sigma2, s[:pos] ** 2)) ** 2 - 4 * L * M * sigma2 ** 2 / s[:pos] ** 4))

    # Computation of the posterior
    post = {}
    post['ma'] = np.zeros(H) 
    post['mb'] = np.zeros(H)
    post['sa2'] = np.zeros(H) 
    post['sb2'] = np.zeros(H) 
    post['cacb'] = np.zeros(H)  

    tau = np.multiply(d, s[:pos]) / (M * sigma2)
    delta = np.multiply(np.sqrt(np.divide(M * d, L * s[:pos])), 1 + alpha / tau)

    post['ma'][:pos] = np.sqrt(np.multiply(d, delta))
    post['mb'][:pos] = np.sqrt(np.divide(d, delta))
    post['sa2'][:pos] = np.divide(sigma2 * delta, s[:pos])
    post['sb2'][:pos] = np.divide(sigma2, np.multiply(delta, s[:pos]))
    post['cacb'][:pos] = np.sqrt(np.multiply(d, s[:pos]) / (L * M))
    post['sigma2'] = sigma2
    post['F'] = 0.5 * (L * M * np.log(2 * np.pi * sigma2) + (residual + np.sum(s ** 2)) / sigma2 
                       + np.sum(M * np.log(tau + 1) + L * np.log(tau / alpha + 1) - M * tau))

    return U[:, :pos], np.diag(d), V[:, :pos], post

def EVBsigma2(sigma2, L, M, s, residual, xubar):
    """
    Compute the objective function for optimizing the noise variance (sigma2) in Empirical Variational Bayesian Matrix Factorization (EVBMF).

    Args:
        sigma2 (float): Noise variance to optimize.
        L (int): Number of rows in the input matrix.
        M (int): Number of columns in the input matrix.
        s (numpy.ndarray): Singular values of the matrix to factorize.
        residual (float): Residual sum of squares from truncated singular values.
        xubar (float): Threshold for singular values based on prior knowledge.

    Returns:
        float: Objective function value for the given noise variance.
    """
    H = len(s)

    alpha = L / M
    x = s ** 2 / (M * sigma2) 

    z1 = x[x > xubar]
    z2 = x[x <= xubar]
    tau_z1 = tau(z1, alpha)

    term1 = np.sum(z2 - np.log(z2))
    term2 = np.sum(z1 - tau_z1)
    term3 = np.sum(np.log(np.divide(tau_z1 + 1, z1)))
    term4 = alpha * np.sum(np.log(tau_z1 / alpha + 1))
    
    obj = term1 + term2 + term3 + term4 + residual / (M * sigma2) + (L - H) * np.log(sigma2)

    return obj

def phi0(x):
    """
    Calculate the phi0 function for EVBMF.

    Args:
        x (float or numpy.ndarray): Input value(s).

    Returns:
        float or numpy.ndarray: Result of the phi0 calculation.
    """
    return x - np.log(x)

def phi1(x, alpha):
    """
    Calculate the phi1 function for EVBMF.

    Args:
        x (float or numpy.ndarray): Input value(s).
        alpha (float): Aspect ratio of the input matrix (L / M).

    Returns:
        float or numpy.ndarray: Result of the phi1 calculation.
    """
    return np.log(tau(x, alpha) + 1) + alpha * np.log(tau(x, alpha) / alpha + 1) - tau(x, alpha)

def tau(x, alpha):
    """
    Compute the tau function, which is part of the EVBMF framework.

    Args:
        x (float or numpy.ndarray): Input value(s).
        alpha (float): Aspect ratio of the input matrix (L / M).

    Returns:
        float or numpy.ndarray: Computed tau values.
    """
    return 0.5 * (x - (1 + alpha) + np.sqrt((x - (1 + alpha)) ** 2 - 4 * alpha))


from tensorly.decomposition import partial_tucker
import tensorly as tl
import torch
import torch.nn as nn
import numpy as np

def tucker_decomposition_conv_layer(layer, ranks):
    """ 
    Gets a convolutional layer, applies Tucker decomposition and returns the set of smaller layers.

    Args:
        layer (torch layer): convolutional layer to decompose
        ranks (tuple): Ranks of the Tucker decomposition

    Returns 
        nn.Sequential: object with the Tucker decomposition.
    """
    tensors, _ = \
        partial_tucker(tl.tensor(layer.weight.data), \
            ranks, modes=[0, 1], init='svd')
    core, [last, first] = tensors
    
    # A pointwise convolution that reduces the channels from S to R3
    first_layer = torch.nn.Conv2d(in_channels=first.shape[0], \
            out_channels=first.shape[1], kernel_size=1,
            stride=1, padding=0, dilation=layer.dilation, bias=False)

    # A regular 2D convolution layer with R3 input channels and R3 output channels
    core_layer = torch.nn.Conv2d(in_channels=core.shape[1], \
            out_channels=core.shape[0], kernel_size=layer.kernel_size,
            stride=layer.stride, padding=layer.padding, dilation=layer.dilation,
            bias=False)

    # A pointwise convolution that increases the channels from R4 to T
    last_layer = torch.nn.Conv2d(in_channels=last.shape[1], \
                                 out_channels=last.shape[0], kernel_size=1, stride=1,
                                 padding=0, dilation=layer.dilation, bias=True)

    if layer.bias is not None:
        last_layer.bias.data = layer.bias.data

    first_layer.weight.data = \
        torch.transpose(torch.tensor(first), 1, 0).unsqueeze(-1).unsqueeze(-1)
    last_layer.weight.data = torch.tensor(last).unsqueeze(-1).unsqueeze(-1)
    core_layer.weight.data = torch.tensor(core)

    new_layers = [first_layer, core_layer, last_layer]
    return nn.Sequential(*new_layers)

def count_params(network):
    '''
    Counts the number of parameters in the neural network

    Args:
        network (torch.nn): Neural network model
        
    Returns:
        float: Number of training parameters
    '''
    return np.sum(np.prod(p.size()) for p in network.parameters())




import collections
import itertools

from qiskit import QuantumCircuit, QuantumRegister, ClassicalRegister, transpile
from qiskit_aer import AerSimulator
from qiskit.quantum_info import Pauli, random_unitary, Operator, SparsePauliOp
from qiskit.circuit.library import PauliEvolutionGate
from qiskit_ibm_runtime import Session, EstimatorV2 as Estimator, SamplerV2 as Sampler

import numpy as np
import pickle
from random import sample
import torch
import string
from .utils import roll_numpy, roll_torch
import torch.nn.functional as F

class QuantumFiltersBase():
    """
    Base class for quantum-inspired filters applied to 2D and 3D images.

    This class implements different quantum encoding schemes (`FRQI` and `angle`)
    and quantum transformations using random quantum circuits or Ising models.

    Attributes:
        n_dimensions (int): Number of spatial dimensions (2D or 3D).
        shape (tuple): Shape of the filter (e.g., (4,4) for 2D, (4,4,4) for 3D).
        stride (float): Stride used when applying the quantum filter.
        encoding (str): Encoding method ('frqi' or 'angle').
        shots (int): Number of shots for Qiskit experiments.
        backend (str): Execution backend ('torch' or 'aer_simulator').
        nqbits (int): Number of qubits used in the quantum circuit.
        device (torch.device, optional): PyTorch device ('cpu' or 'cuda'), only if using `torch` backend.
        unitaries_list (list): List of unitary transformations used in quantum filtering.
        num_filters (int): Number of quantum filters applied per feature.
        num_features (int): Number of features in the input data.
        gates_name (str): Gate set used for quantum transformation ('G1', 'G2', 'G3', or 'Ising').
        info (dict): Metadata containing circuit details such as depth and nonlocal gate count.
    """
    def __init__(
            self, n_dimensions: int, shape: tuple, stride: float, encoding: str,
            shots: int, backend: str):
        """
        Initializes the quantum filter base class.

        Args:
            n_dimensions (int): Number of spatial dimensions (2D or 3D).
            shape (tuple): Shape of the filter.
            stride (float): Stride used when applying the quantum filter.
            encoding (str): Encoding method ('frqi' or 'angle').
            shots (int): Number of shots for Qiskit experiments.
            backend (str): Execution backend ('torch' or 'aer_simulator').

        Raises:
            ValueError: If an unsupported encoding method or backend is provided.
        """

        self.n_dimensions = n_dimensions

        # Calculate the number of qubits
        if encoding=='angle':
            self.nqbits =shape[0]**n_dimensions
        elif encoding=='frqi':    
            self.nqbits = int(np.ceil(np.log2(shape[0]**n_dimensions)) + 1)
        else:
            raise ValueError('Data encoding method not implemented. The available encoding methods are frqi and angle.')
        self.encoding = encoding

        self.shape = shape
        self.stride = stride
        self.shots = shots
        self.info = {}

        self.backend = backend
        if self.backend == 'torch':
            # set CUDA for PyTorch
            use_cuda = torch.cuda.is_available()
            if use_cuda:
                self.device = torch.device("cuda:0")
                torch.cuda.set_device(int("cuda:0".split(':')[1]))
            else:
                self.device = torch.device("cpu")
        elif backend=='aer_simulator':
            self.backend = AerSimulator()
        elif type(backend) == str:
            raise ValueError('The only valid backend names are "torch" and "aer_simulator". You can also provide a Qiskit Backend directly.')

        # Initialise
        self.unitaries_list = []
        self.num_filters = 0
        self.num_features = 0

    gate_names = {
        "G1": ["CNOT", "H", "X"],
        "G2": ["CNOT", "H", "S"],
        "G3": ["CNOT", "H", "T"],
    }

    def _select_gates_qubits(self, gates_name, num_gates):
        """
        Selects random quantum gates and corresponding qubits.

        Args:
            gates_name (str): Name of the gate family ('G1', 'G2', 'G3', or 'Ising').
            num_gates (int): Number of gates to apply.

        Returns:
            tuple: 
                - List of quantum gates used in the circuit.
                - List of qubits where each gate is applied.

        Raises:
            ValueError: If an invalid gate set name is provided.
        """
        # Select random gate
        gates = self.gate_names[gates_name]
        gates_set = [sample(gates, 1)[0] for _ in range(num_gates)]

        qubit_idx = list(range(self.nqbits))

        def _add_qubit_set(gate):
            if gate == 'CNOT':
                # Select qubit 1 and 2 (different qubits)
                qbit1 = sample(qubit_idx, 1)[0]
                qubit_idx2 = qubit_idx.copy()
                qubit_idx2.remove(qbit1)
                qbit2 = sample(qubit_idx2, 1)[0]
                return [qbit1, qbit2]
            else:
                # Select qubit
                qbit = sample(qubit_idx, 1)[0]
                return [qbit]

        # Store qubit list of applied gates
        qubits_set = [
            _add_qubit_set(gate)
            for gate in gates_set
        ]
        return gates_set, qubits_set

    def _apply_G_gates(self, qc, gates_set, qubits_set, measure=False):
        """
        Selects a set of random qubits and random gates from the G1, G2 or G3
        family, and stores them in a list.
        These gates form the random quantum circuit.

        Args:
            qc (QuantumCircuit): quantum circuit
            gates_set (list): List of quantum gates that form the quantum
                reservoir
            qubits_set (list): List of qubits to which the quantum gates are
                applied to
            measure (bool): measure all the qubits (except the ancilla) at the
                end of the circuit
        """
        # Apply random gates to random qubits
        for gate, qubit in zip(gates_set, qubits_set):

            if gate == 'CNOT':  # For 2-qubit gates
                # Select qubit 1 and 2 (different qubits)
                qbit1, qbit2 = qubit
                # Apply gate to qubits
                qc.cx(qbit1, qbit2)

            else:  # For 1-qubit gates
                # Select qubit
                qbit = qubit[0]

                # Apply gate
                if gate == 'X':
                    qc.x(qbit)
                elif gate == 'S':
                    qc.s(qbit)
                elif gate == 'H':
                    qc.h(qbit)
                elif gate == 'T':
                    qc.t(qbit)
                else:
                    raise ValueError("Unknown quantum gate:", gate)
        if measure:
            qc.measure(list(range(self.nqbits-1)), list(range(self.nqbits-1)))

    def _select_Ising_coefficinets(self):
        """
        Defines random Ising coefficients.

        Returns:
            Js (float): J coefficients
            h (float): h coefficient
        """
        # Define parameters Ising model
        Js = np.random.normal(0.75, 0.1)
        h_over_Js = 0.1
        h = Js*h_over_Js
        return Js, h

    
    def _apply_ising_gates(self,Js=None,h=None, t: float = 10):
        """
        Applies quantum evolution under the transverse field Ising model.
        The parameters of the Ising model are chosen according to
        R. Martínez-Peña et al. (2021).
        Js ~ N(0.75, 0.1), Jij ~ U(-Js/2, Js/2) and hi = h such that h/Js = 0.1

        Args:
            Js (float): J coefficients
            h (float): h coefficient
            t (float): time evolution

        Returns:
            (QuantumCircuit): quantum circuit representing the evolution of
            the Ising model
        """
        # Define parameters Ising model
        if Js is None or h is None:
            print('Generating new coefficients')
            Js = np.random.normal(0.75, 0.1)
            h_over_Js = 0.1
            h = Js*h_over_Js

        # Get list of all qubits and pairs of qubits
        qubit_idxs = list(range(self.nqbits))

        # Define the Ising model operators
        pauli_op = 0
        name_gate = 'I' * self.nqbits
        for q1_idx, q2_idx in itertools.combinations(qubit_idxs, 2):
            # Interaction operators
            name = \
                name_gate[:q1_idx] + 'Z' + \
                name_gate[q1_idx+1:q2_idx] + \
                'Z' + name_gate[q2_idx+1:]
            coef = np.random.uniform(-Js/2, Js/2)
            pauli_op += coef*SparsePauliOp(name)

        for q_idx in qubit_idxs:  # Single qubit operators
            name = name_gate[:q_idx] + 'X' + name_gate[(q_idx+1):]
            coef = h
            pauli_op += coef*SparsePauliOp(name)

        # Time evolution circuit
        evolution_gate = PauliEvolutionGate(pauli_op, time=t)
        qc_ham = QuantumCircuit(self.nqbits)
        qc_ham.append(evolution_gate, list(range(self.nqbits)))

        return qc_ham

    def generate_unitaries(
            self, gates_name='G3', num_gates=300, num_filters=3,
            num_features=19, unitaries_file_name='unitaries.pickle',
            save=False):
        """
        Generates the quantum unitaries that represent the quantum reservoirs
        (random quantum circuits)
        This function is only used with Pytorch backend

        Args:
            gates_name (str): name of the family of quantum gates. Either G1,
            G2, G3 or Ising
            num_gates (int): depth of the quantum circuits
            num_filters (int): Number of filters to apply to each feature
            num_features (int): Number of features of the data
            unitaries_file_name (str): name of the file containing unitary list
                (only needed if save=True)
            save (bool): Whether the generated unitaries are saved to a file
        """
        if self.backend != 'torch':
            raise ValueError(
                "This function is only callable with the 'torch' backend")

        self.unitaries_list = []
        self.num_filters = num_filters
        self.num_features = num_features
        self.gates_name = gates_name
        self.info['num_features'] = self.num_features*self.num_filters
        self.info['nqbits'] = self.nqbits
        self.info['depth'] = []
        self.info['num_nonlocal_gates'] = []
        for i in range(num_filters):
            U_list = []
            for _ in range(num_features):
                # Store circuit parameters
                self.num_gates = num_gates
                if gates_name == 'random':
                    U = torch.tensor(random_unitary(2**self.nqbits).data)
                    U_list.append(U)
                    self.info['family_type'] = 'Universal (Haar)'
                    self.info['nqbits'] = self.nqbits
                else:
                    if gates_name == 'Ising':
                        # If the quantum reservoir is the Ising model
                        qc = self._apply_ising_gates()
                        self.info['family_type'] = 'Universal (Ising)'
                        self.info['depth'].append(qc.depth())
                        self.info['num_nonlocal_gates'].append(qc.num_nonlocal_gates())
                    else:
                        # Otherwise, if it is one of the G families
                        self.info['family_type'] = 'Clifford (subgroup)' if gates_name == "G1" else 'Clifford group' if gates_name=='G2' else 'Universal'
                        gates_set, qubits_set = \
                            self._select_gates_qubits(gates_name, num_gates)
                        qc = QuantumCircuit(self.nqbits)
                        # Random quantum circuit with gates
                        self._apply_G_gates(qc, gates_set, qubits_set)
                        self.info['depth'].append(qc.depth())
                        self.info['num_nonlocal_gates'].append(qc.num_nonlocal_gates())
                    # Get unitary
                    U = Operator(qc).data 
                    U_list.append(U)
            self.unitaries_list.append(U_list)
        # Save unitaries to file
        if save:
            with open(unitaries_file_name, 'wb') as f:
                pickle.dump(
                    self.unitaries_list, f, protocol=pickle.HIGHEST_PROTOCOL)

    def _check_file_names(self, saved_gates_filename, saved_qubits_filename):
        """
        Checks if file names are defined, otherwise it creates them.

        Args:
            saved_gates_filename (str/None): Name of the gates file
            saved_qubits_filename (str/None): Name of the qubits file

        Returns:
            saved_gates_filename (str): Name of the gates file
            saved_qubits_filename (str): Name of the qubits file
        """

        saved_gates_filename = \
            saved_gates_filename or f'gates_list_{self.n_dimensions}D.pickle'
        saved_qubits_filename = \
            saved_qubits_filename or f'qubits_list_{self.n_dimensions}D.pickle'

        return saved_gates_filename, saved_qubits_filename

    def generate_reservoirs(self, gates_name='G3', num_gates=300, num_filters=3,
            num_features=19, save=False, filename='reservoir.pickle'):
        '''
        Generate sets of random quantum gates and their associated qubits and
        saves them for Qiskit backends, and unitary tensors for torch backend.

        Args:
            gates_name (str): name of the family of quantum gates. Either G1,
            G2, G3 or Ising
            num_gates (int): depth of the quantum circuits
            num_filters (int): Number of filters to apply to each feature
            num_features (int): Number of features of the data
            save (bool): Save the gates and qubits to pickle files
            filename (str): File name for saved gates or unitaries
        '''
        if self.backend == 'torch':
            self.generate_unitaries( gates_name, num_gates, num_filters,
                    num_features, save = save, unitaries_file_name = filename)
        else:
            if filename is not None:
                saved_gates_filename = 'gates_' + filename
                saved_qubits_filename = 'qubits_' + filename
            else:
                saved_gates_filename = None
                saved_qubits_filename = None
            self.generate_qc(gates_name, num_gates, num_filters,
            num_features, save,saved_gates_filename, saved_qubits_filename)


    def generate_qc(
            self, gates_name='G3', num_gates=300, num_filters=3,
            num_features=19, save=False,
            saved_gates_filename='gates.pickle', saved_qubits_filename='qubits.pickle'):
        '''
        Generate sets of random quantum gates and their associated qubits and
        saves them. This function is only used with Qiskit backends.

        Args:
            gates_name (str): name of the family of quantum gates. Either G1,
            G2, G3 or Ising
            num_gates (int): depth of the quantum circuits
            num_filters (int): Number of filters to apply to each feature
            num_features (int): Number of features of the data
            save (bool): Save the gates and qubits to pickle files
            saved_gates_filename (str): File name for saved gates set
            saved_qubits_filename (str): File name for saved qubit set
        '''
        if self.backend == 'torch':
            raise ValueError(
                "This function is only callable with the Qiskit backends")

        saved_gates_filename, saved_qubits_filename = \
            self._check_file_names(saved_gates_filename, saved_qubits_filename)

        self.gates_set_list = []
        self.qubits_set_list = []
        # Store circuit parameters
        self.gates_name = gates_name
        self.num_filters = num_filters
        self.num_features = num_features
        self.info['family_type'] = 'Clifford (subgroup)' if gates_name == "G1" else 'Clifford group' if gates_name=='G2' else 'Universal' if gates_name=='G3' else 'Universal (Ising)'
        self.info['num_circuits'] = self.num_features*self.num_filters
        self.info['nqbits'] = self.nqbits
        self.info['depth'] = []
        self.info['num_nonlocal_gates'] = []
        self.num_gates = num_gates

        if self.gates_name=='Ising':
            self.J_list = []
            self.h_list = []
            for _ in range(num_filters):
                J_list = []
                h_list = []
                for _ in range(num_features):
                    Js, h = self._select_Ising_coefficinets()   
                    J_list.append(Js)
                    h_list.append(h)
                self.J_list.append(J_list)
                self.h_list.append(h_list)
            if save:
                # Save Ising coefficients to file
                with open(saved_gates_filename, 'wb') as f:
                    pickle.dump(
                        self.J_list, f, protocol=pickle.HIGHEST_PROTOCOL)
                with open(saved_qubits_filename, 'wb') as f:
                    pickle.dump(
                        self.h_list, f, protocol=pickle.HIGHEST_PROTOCOL)             
        else:
            for _ in range(num_filters):
                # Random quantum circuit for each filter
                gates_list = []
                qubits_list = []
                for _ in range(num_features):
                    # Random quantum circuit for each feature
                    gates_set, qubits_set = \
                        self._select_gates_qubits(gates_name, num_gates)
                    gates_list.append(gates_set)
                    qubits_list.append(qubits_set)

                self.gates_set_list.append(gates_list)
                self.qubits_set_list.append(qubits_list)

            # Save gates and qubits to file
            if save:
                with open(saved_gates_filename, 'wb') as f:
                    pickle.dump(
                        self.gates_set_list, f, protocol=pickle.HIGHEST_PROTOCOL)
                with open(saved_qubits_filename, 'wb') as f:
                    pickle.dump(
                        self.qubits_set_list, f, protocol=pickle.HIGHEST_PROTOCOL)

    def load_reservoirs(self, gates_name='G3', filename = None):
        '''
        Load set of quantum gates and qubits for Qiskit backends, or unitaries 
        for pytorch backends.

        Args:
            gates_name (str): name of the family of quantum gates. Either G1,
            G2, G3 or Ising
            filename (str): File name for saved gates set or unitaries
        '''
        if self.backend=='torch':
            self.load_unitaries(filename)
        else:
            self.load_gates(gates_name, 'gates_' + filename, 'qubits_' + filename)

    def load_gates(
            self, gates_name='G3',
            saved_gates_filename=None, saved_qubits_filename=None):
        '''
        Load set of quantum gates and qubits. This function is only used with
        Qiskit backends.

        Args:
            gates_name (str): name of the family of quantum gates. Either G1,
            G2, G3 or Ising
            saved_gates_filename (str): File name for saved gates set
            saved_qubits_filename (str): File name for saved qubit set
        '''

        saved_gates_filename, saved_qubits_filename = \
            self._check_file_names(saved_gates_filename, saved_qubits_filename)

        if self.backend == 'torch':
            raise ValueError(
                "This function is only callable whith the Qiskit backends")

        with open(saved_gates_filename, 'rb') as f:
            self.gates_set_list = pickle.load(f)
        with open(saved_qubits_filename, 'rb') as f:
            self.qubits_set_list = pickle.load(f)

        self.gates_name = gates_name
        # Store circuit parameters
        self.num_filters = len(self.gates_set_list)
        self.num_features = len(self.gates_set_list[0])
        
        if self.gates_name=='Ising':
            self.J_list = self.gates_set_list
            self.h_list = self.qubits_set_list
        else:
            self.num_gates = len(self.gates_set_list[0][0])
            nqbits = max(max(max(max(self.qubits_set_list)))) + 1
            if nqbits != self.nqbits:
                raise ValueError(
                    'Incorrect number of qubits of the loaded quantum circuits')


    def load_unitaries(self, file_name):
        '''
        Loads the unitaries.  This function is only used with Pytorch backend.

        Args:
            file_name (str): File name for unitaries
        '''
        if self.backend != 'torch':
            raise ValueError(
                "This function is only callable with the 'torch' backend")

        with open(file_name, 'rb') as f:
            self.unitaries_list = pickle.load(f)

        # Store circuit parameters
        self.num_filters = len(self.unitaries_list)
        self.num_features = len(self.unitaries_list[0])
        nqbits = int(np.ceil(np.log2(self.unitaries_list[0][0].shape[0])))

        if nqbits != self.nqbits:
            raise ValueError(
                'Incorrect number of qubits of the loaded quantum circuits')

    def _FRQI_encoding(self, box):
        '''
        Flexible representation of Quantum Images encoding. Takes an (nxn) or
        (nxnxn) box and returns the encoded quantum circuit.
        This function is only used with Qiskit backends.

        Args:
            box (np.array): (nxn) or (nxnxn) box to be encoded

        Returns:
            qc (QuantumCircuit): quantum circuit with FRQI encoding
            initial_state (np.array): Initial state array
        '''
        # check if the box has the correct shape
        assert box.shape == self.shape

        # Calculate binary string of basis qubits
        qbit_list = list(itertools.product([0, 1], repeat=self.nqbits-1))
        coefs = {}

        def _add_coeffs(theta, idx):
            q_str = ''.join([str(q) for q in qbit_list[idx]])
            coefs[q_str + '0'] = np.cos(theta) / \
                np.sqrt(self.shape[0]**self.n_dimensions)
            coefs[q_str + '1'] = np.sin(theta) / \
                np.sqrt(self.shape[0]**self.n_dimensions)

        l_idx = 0
        for i in range(box.shape[0]):
            for j in range(box.shape[1]):
                if self.n_dimensions == 2:
                    _add_coeffs(box[i, j], l_idx)
                    l_idx += 1
                elif self.n_dimensions == 3:
                    for k in range(box.shape[2]):
                        _add_coeffs(box[i, j, k], l_idx)
                        l_idx += 1

        # store them in a dictionary and sort it by keys
        coefs = list(collections.OrderedDict(sorted(coefs.items())).values())

        # Define quantum circuit
        cr = ClassicalRegister(self.nqbits-1)
        qr = QuantumRegister(self.nqbits)
        qc = QuantumCircuit(qr, cr)

        # Calculate initial state
        initial_state = np.zeros(2**self.nqbits)
        coefs = coefs/np.sqrt(np.sum(np.square(coefs)))
        initial_state[:coefs.shape[0]] = coefs
        qc.initialize(initial_state, list(range(self.nqbits)))
        return qc, initial_state

    def _angle_encoding(self, box):
        '''
        Angle encoding for images. Takes an (nxn) or
        (nxnxn) box and returns the encoded quantum circuit.
        This function is only used with Qiskit backends.

        Args:
            box (np.array): (nxn) or (nxnxn) box to be encoded

        Returns:
            qc (QuantumCircuit): quantum circuit with angle encoding
        '''
        # check if the box has the correct shape
        assert box.shape == self.shape
        # Define quantum circuit
        cr = ClassicalRegister(self.nqbits)
        qr = QuantumRegister(self.nqbits)
        qc = QuantumCircuit(qr, cr)
        flattened_box = box.flatten()
        for (i,angle) in enumerate(flattened_box):
            qc.rx(angle,i)
        
        return qc


    def _scale_data(self, data):
        """
        Scale the data to [0, pi/2) (each feature is scaled separately)

        Args:
            data (tensor): input data, shape (n_samples, num_features, N,N,N)

        Returns:
            (tensor): scaled data, shape (n_samples, num_features, N,N,N)
        """
        new_min, new_max = 0, np.pi/2
        for i in range(data.shape[0]):  # Scale each feature independently
            for j in range(data.shape[1]):

                if self.n_dimensions == 2:
                    data_slice = data[i, j, :, :]
                elif self.n_dimensions == 3:
                    data_slice = data[i, j, :, :, :]

                v_min, v_max = data_slice.min(), data_slice.max()
                if v_min < v_max:
                    data_slice = (data_slice - v_min) / \
                        (v_max - v_min)*(new_max - new_min) \
                        + new_min
        return data

    def _run_boxQiskit(self, box, gates_set, qubits_set):
        '''
        For a given box of the image runs the FRQI encoding + random quantum
        circuit.  This function is only used with Qiskit backends.

        Args:
            box (np.array): shape (n,n). Input data
            gates_set (list): Set of random quantum gates
            qubits_set (list): List of qubits to apply the quantum gates

        Returns:
            np.array: Quantum reservoir for a box in the image
        '''
        
        if self.encoding=='frqi':
            # FRQI encoding
            qc, _ = self._FRQI_encoding(box)
        elif self.encoding=='angle':
            # Angle encoding
            qc = self._angle_encoding(box)
    
        # Random quantum circuit
        if self.gates_name == 'Ising':
            qc_Ising = self._apply_ising_gates(gates_set, qubits_set)
            qc = qc.compose(qc_Ising, qubits=range(qc.num_qubits))
            #qc += qc_Ising
        else:
            self._apply_G_gates(qc, gates_set, qubits_set, measure=True)

        self.info['depth'].append(qc.depth())
        self.info['num_nonlocal_gates'].append(qc.num_nonlocal_gates())
        # Measure
        qc.measure_all()
        # Get counts
        is_simulator = self.backend.configuration().simulator # Check if we're using a simulator or quantum hardware
        with Session(backend=self.backend) as session: # Create a session with such backend
            sampler = Sampler(mode=session) if self.encoding =='frqi' else Estimator(mode=session)
            sampler.options.default_shots = self.shots
            qc = transpile(qc, self.backend)

            # If running on real hardware (not a simulator), apply error mitigation
            if not is_simulator:
                sampler.options.dynamical_decoupling.enable = True
                sampler.options.dynamical_decoupling.sequence_type = "XY4"
                sampler.options.twirling.enable_gates = True
                sampler.options.twirling.num_randomizations = "auto"

            if self.encoding=='frqi':                
                pub = (qc,) # Prepare input
                job = sampler.run([pub]) 
                counts = job.result()[0].data.meas.get_counts()
                # Calculate binary string of basis qubits
                new_counts = [
                    counts.get(''.join([str(value) for value in qbits]), 0)
                    for qbits in itertools.product([0, 1], repeat=qc.num_qubits)#self.nqbits-1)
                ]
                #print(counts, new_counts)
                new_counts /= np.sum(new_counts)
                return new_counts
            
            else:
                exp_vals = []
                # Get operators
                for index in range(self.nqbits):
                    name = "I"*self.nqbits
                    name=name[:index] + 'Z' + name[index + 1:]
                    op = SparsePauliOp(name)
                    # Run expected value
                    isa_hamiltonian = op.apply_layout(qc.layout)
                    pub = (qc, isa_hamiltonian,)
                    job = sampler.run([pub])
                    results = job.result()[0]
                    exp_vals.append(results.data.evs)               

                return np.array(exp_vals)

    def _run_filter_qiskit(self, data, gates_set, qubits_set, tol=1e-6):
        '''
        Runs the quantum circuit for one feature of the data.  This function
        is only used with Qiskit backends.

        Args:
            data (np.array): shape (N,N), channel of the image
            gates_set (list): Set of random quantum gates
            qubits_set (list): List of qubits to apply the quantum gates
            scale (bool): Scale the input image

        Returns:
            np.array: Quantum reservoir for the whole image
        '''

        size = self.shape[0]

        transpose_idxs = {
            2: [0, 2, 1, 3],
            3: [0, 3, 1, 4, 2, 5]
        }

        # Get boxes from data
        strided_window, shape = roll_numpy(
            data, np.zeros(self.shape),
            dx=size*self.stride, dy=size*self.stride,
            dz=size*self.stride if self.n_dimensions == 3 else None
        )
        self.shape_windows = shape
        windows = strided_window.reshape((-1,) + (size,) * self.n_dimensions)

        def _run_per_box(box):
            if np.sum(np.abs(box)) > tol:
                # Run the QC if the box is non-zero valued
                result = np.array(
                    self._run_boxQiskit(box, gates_set, qubits_set))
                return result[:size**self.n_dimensions].reshape(self.shape)
            else:
                # Return zeros if the box is all zero-valued
                return np.array(
                    [0] * size**self.n_dimensions
                ).reshape(self.shape)

        results = [
            _run_per_box(windows[i])
            for i in range(windows.shape[0])  # Every box of the image
        ]

        fin_shape = int(data.shape[-1]/self.stride)
        fin_shape_parameters = (fin_shape,) * self.n_dimensions

        # Reshape the results in the original shape
        results = np.array(results).reshape(self.shape_windows)
        results = results.transpose(transpose_idxs[self.n_dimensions]) \
            .reshape(fin_shape_parameters)  # Transpose to original shape

        # Apply zero-mask
        mask = np.ones(self.shape_windows)

        # Reshape to rolling windows shape
        windows_reshape = windows.reshape(self.shape_windows)

        # Find zero-valued values of the original data
        mask[np.abs(windows_reshape) < tol] = 0

        mask = mask.reshape(self.shape_windows) \
            .transpose(transpose_idxs[self.n_dimensions]) \
            .reshape(fin_shape_parameters)  # Transpose to original shape
        results = results.astype('float64')
        results *= mask  # Apply mask

        return results

    def _run_filter_torch_angle(self, data, tol, U):
        """ 
        Runs quantm filter with angle encoding.
        This function is only used with Pytorch backend.

        Args:
            data (tensor): input data (one feature), shape (num_samples, N,N,N)
            tol (float): tolerance for the masking matrix. All values from the
            original data which are smaller than the tolerance are set to 0
            U (tensor): Unitary matrix representing the random quantum circuit

        Returns:
            (tensor): quantum filter applied to data
        """
        size = self.shape[0]
        data = data.reshape(-1,1,data.shape[-1], data.shape[-1])
        #print('data', data.shape)
        # Get measurement operator
        Zs = []
        for index in range(self.nqbits):
            name = "I"*self.nqbits
            name=name[:index] + 'Z' + name[index + 1:]
            Zs.append(torch.tensor(Pauli(name).to_matrix()))

        # 1. Get rolling windows
        reshape_idxs = (-1,) + (size,) * 2
        #print('reshape_idxs', reshape_idxs)
        strided_window = torch.nn.functional.unfold(data, (size,size), stride=self.stride)
        #print('strided_window', strided_window.shape)
        windows = strided_window.reshape(*reshape_idxs)
        #print('windows', windows.shape)
        # 2. Get initial state (rotation for each qubit)
        v1 = torch.tensor([[1, 0]])
        v2 = torch.tensor([[0, 1]])
        flat_block = windows.reshape(-1,size*size)

        cos_block = torch.cos(flat_block)
        sin_block = torch.sin(flat_block)
        prod = cos_block[:,:, None] * v1 + sin_block[:,:, None] * v2
        # 2. Get the tensor product of the [cos(theta),sin(theta)] vectors to create the final state
        letters = list(string.ascii_lowercase)
        l1 = letters[0]
        # Create the formula string
        formula = f"{','.join([l1 + letters[i] for i in range(1, self.nqbits + 1)])}->{''.join(letters[:self.nqbits+1])}"
        # Create the list of products using list comprehension
        prods = [prod[:, i] for i in range(self.nqbits)]
        all_prod = torch.einsum(formula,*prods).reshape(-1,2**self.nqbits).type(torch.complex128)
        # 3. Apply unitary matrix
        apply_U = (U @ all_prod.T).T
        # 4. Calculate expected values
        exps = None
        for Z in Zs:
            exp = ((apply_U.conj() @ Z) * apply_U).sum(dim=1).real.reshape(-1,1)
            if exps is None:
                exps = exp
            else:
                exps =torch.concat((exps, exp), axis=1)
        exps = exps.reshape(1,1,-1,size,size)

        # 5. Reshape to original image
        output_strided = exps.reshape(strided_window.shape)
        quantum_filter = torch.nn.functional.fold(output_strided, output_size = (data.shape[-1], data.shape[-1]),
                                                   kernel_size=(size,size), stride=(self.stride, self.stride))
        quantum_filter = quantum_filter.reshape(data.shape[0], data.shape[-1],data.shape[-1])
        #print('quantum_filter', quantum_filter.shape)
        return quantum_filter

                
    def _run_filter_torch_frqi(self, data, tol, U):
        """
        Runs a quantum filter to a feature from the data samples with Flexible Representation of Quantum Images.
        This function is only used with Pytorch backend.

        Args:
            data (tensor): input data (one feature), shape (num_samples, N,N,N)
            tol (float): tolerance for the masking matrix. All values from the
            original data which are smaller than the tolerance are set to 0
            U (tensor): Unitary matrix representing the random quantum circuit

        Returns:
            (tensor): quantum filter applied to data
        """
        size = self.shape[0]
        reshape_idxs = (-1,) + (size,) * self.n_dimensions

        # Set parameters based on number of dimensions
        if self.n_dimensions == 2:
            einsum_equation = 'ijkl->ijk'
            permutation_idxs = [0, 1, 3, 2, 4]
            dz = None
        elif self.n_dimensions == 3:
            einsum_equation = 'ijklm->ijkl'
            permutation_idxs = [0, 1, 4, 2, 5, 3, 6]
            dz = size*self.stride

        # 0. Get rolling windows
        strided_window, shape = roll_torch(
            data, torch.zeros(self.shape),
            dx=size*self.stride, dy=size*self.stride, dz=dz
        )
        self.shape_windows = shape
        windows = strided_window.reshape(*reshape_idxs).to(self.device)

        # 1. Flexible representation of quantum images.
        # Apply cos(theta)|0> + sin(theta)|1> transformation
        v1 = torch.tensor([1, 0]).to(self.device)
        v2 = torch.tensor([0, 1]).to(self.device)
        sq = torch.tensor([size**self.n_dimensions]).to(self.device)

        frqi = (
            torch.kron(torch.cos(windows), v1) +
            torch.kron(torch.sin(windows), v2)
        ) / torch.sqrt(sq)

        # 2. Flatten the last dimension to create an initial state
        flatten = frqi.reshape(-1, 2**(self.nqbits)).type(torch.complex128)
        # 3. Apply U rotation to each state
        apply_U = (U @ flatten.T).reshape(reshape_idxs + (2,))

        # 4. Partial trace to remove the extra qubit
        partial_trace = torch.einsum(einsum_equation, apply_U)

        # 5. Reshape back to window view
        partial_trace_reshaped = partial_trace.reshape(self.shape_windows)

        # 6. Reshape to original shape, taking into account transposition
        fin_shape = int(data.shape[-1]/self.stride)
        new_reshape_idxs = \
            (self.num_samples,) + (fin_shape,) * self.n_dimensions
        output = partial_trace_reshaped.permute(
            permutation_idxs).reshape(new_reshape_idxs)

        # 7. Get the probability for each state (modulus of the final state)
        output_probability = (output.conj()*output).real

        # 8. Create a mask to map to zero all zero values of the original image
        mask = torch.ones(windows.shape)
        mask[torch.abs(windows) < tol] = 0
        mask = mask.reshape(self.shape_windows).permute(permutation_idxs) \
            .reshape(new_reshape_idxs).to(self.device)

        quantum_filter = output_probability*mask
        return quantum_filter

    def _run(self, data, tol=1e-6, n_filt=0):
        """
        Runs the quantum filters for all the features

        Args:
            data (tensor): input data (one feature), shape
                (num_samples, num_features, N,N,N)
            tol (float): tolerance for the masking matrix. All values from the
                original data which are smaller than the tolerance are set to 0
            n_filt (int): index of the number of filters

        Returns:
            (tensor): output quantum filter
        """
        # Scale data
        data_scaled = self._scale_data(data)
        self.num_samples = data.shape[0]  # Store number of samples
        fin_shape = int(data.shape[-1]/self.stride)  # Final shape of the data
        
        if self.backend == 'torch':
            if self.encoding=='frqi':
                fun_filter = self._run_filter_torch_frqi
            elif self.encoding=='angle':
                fun_filter = self._run_filter_torch_angle
            
            data_out = torch.zeros(
                (data.shape[0], data.shape[1],) +
                (fin_shape,) * self.n_dimensions
            )

            for i in range(data.shape[1]):  # Run for every feature
                unitary_matrix = torch.tensor(
                    self.unitaries_list[n_filt][i]
                ).to(self.device)

                if self.n_dimensions == 2:
                    data_out[:, i, :, :] = fun_filter(
                        data_scaled[:, i, :, :], tol, unitary_matrix
                    )
                elif self.n_dimensions == 3:
                    data_out[:, i, :, :, :] = fun_filter(
                        data_scaled[:, i, :, :, :], tol, unitary_matrix
                    )
        else:
            data_out = np.zeros(
                (data.shape[0], data.shape[1],) +
                (fin_shape,) * self.n_dimensions
            )

            for i in range(data.shape[0]):
                for j in range(data.shape[1]):
                    gates_set = self.J_list[n_filt][j] if self.gates_name=='Ising' else self.gates_set_list[n_filt][j]
                    qubits_set = self.h_list[n_filt][j] if self.gates_name=='Ising' else self.qubits_set_list[n_filt][j]

                    if self.n_dimensions == 2:
                        data_out[i, j, :, :] = self._run_filter_qiskit(
                            data[i, j, :, :], gates_set, qubits_set, tol
                        )
                    elif self.n_dimensions == 3:
                        data_out[i, j, :, :, :] = self._run_filter_qiskit(
                            data[i, j, :, :, :], gates_set, qubits_set, tol
                        )

        return data_out

    def get_quantum_filters(self, data, tol=1e-6):
        """
        Runs the quantum filters for all features multiple (num_filters) times

        Args:
            data (tensor): input data (one feature),
                shape 2D (num_samples, num_features, N, N),
                3D (num_samples, num_features, N, N, N)
            tol (float): tolerance for the masking matrix. All values from the
                original data which are smaller than the tolerance are set to 0

        Returns:
            (tensor/np.array): output quantum filters
        """
        data = torch.tensor(data) if self.backend=='torch' else np.array(data)
        if len(data.shape) != self.n_dimensions + 2:
            raise ValueError(
                'Incorrect data shape. The data should have shape'
                ' (num_samples, num_features,) + (N,) * number of dimensions')
        all_results = [
            self._run(data, tol, n_filt=i)
            for i in range(self.num_filters)  # Run for each number of filters
        ]

        result_shape = all_results[0].shape
        final_result_shape = [
            result_shape[0],
            result_shape[1]*self.num_filters,
        ] + [
            result_shape[2 + i]
            for i in range(self.n_dimensions)
        ]

        # Reshape final array
        if self.backend == 'torch':
            results_reshaped = torch.zeros(final_result_shape)
            for i, result in enumerate(all_results):
                results_reshaped[
                    :,
                    ((result_shape[1])*i):(result_shape[1])*(i + 1),
                    :,
                    :
                ] = result

        else:
            all_results = np.array(all_results)
            results_reshaped = all_results.reshape(final_result_shape)

        return results_reshaped


class QuantumFilters2D(QuantumFiltersBase):

    """
    Quantum filter for 2D images.

    This class applies a quantum transformation to 2D images using random
    quantum circuits or Ising model evolution.

    Attributes:
        shape (tuple): Shape of the filter (NxN).
        stride (int): Stride used when applying the quantum filter.
        encoding (str): Quantum data encoding method ('frqi' or 'angle').
        shots (int): Number of shots for Qiskit experiments.
        backend (str): Execution backend ('torch' or 'aer_simulator').
    """

    def __init__(
            self, shape: tuple = (4, 4), stride: float = 1,
            encoding: str ='frqi',  shots: int = 4096, backend = 'torch'):
        """
        Initializes a quantum filter for 2D image processing.

        Args:
            shape (tuple, optional): Shape of the filter. Defaults to (4,4).
            stride (float, optional): Stride for the filter. Defaults to 1.
            encoding (str, optional): Quantum encoding ('frqi' or 'angle'). Defaults to 'frqi'.
            shots (int, optional): Number of shots for Qiskit experiments. Defaults to 4096.
            backend (str, optional): Execution backend ('torch' or 'aer_simulator'). Defaults to 'torch'.
        """

        super().__init__(
            n_dimensions=2, shape=shape, stride=stride,
            encoding = encoding, shots=shots, backend=backend)


class QuantumFilters3D(QuantumFiltersBase):

    """
    Quantum filter for 3D images.

    This class applies a quantum transformation to 3D images using random
    quantum circuits or Ising model evolution.

    Attributes:
        shape (tuple): Shape of the filter (NxNxN).
        stride (int): Stride used when applying the quantum filter.
        encoding (str): Quantum data encoding method ('frqi' or 'angle').
        shots (int): Number of shots for Qiskit experiments.
        backend (str): Execution backend ('torch' or 'aer_simulator').
    """

    def __init__(
            self, shape: tuple = (4, 4, 4), stride: float = 1,
            encoding: str ='frqi', shots: int = 4096, backend = 'torch'):
        """
        Initializes a quantum filter for 3D image processing.

        Args:
            shape (tuple, optional): Shape of the filter. Defaults to (4,4,4).
            stride (float, optional): Stride for the filter. Defaults to 1.
            encoding (str, optional): Quantum encoding ('frqi' or 'angle'). Defaults to 'frqi'.
            shots (int, optional): Number of shots for Qiskit experiments. Defaults to 4096.
            backend (str, optional): Execution backend ('torch' or 'aer_simulator'). Defaults to 'torch'.
        """

        super().__init__(
            n_dimensions=3, shape=shape, stride=stride,
            encoding = encoding, shots=shots, backend=backend)
